#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import migraphx, array, tempfile, sys


def test_conv_relu(format):
    p1 = migraphx.parse_onnx("conv_relu_maxpool_test.onnx")
    print(p1)

    s1 = p1.get_output_shapes()[-1]

    with tempfile.NamedTemporaryFile() as t:
        migraphx.save(p1, t.name, format=format)

        p2 = migraphx.load(t.name, format=format)
        print(p2)
        s2 = p2.get_output_shapes()[-1]

        assert s1 == s2
        assert p1.sort() == p2.sort()

def test_save_load_buffer():
    p1 = migraphx.parse_onnx("conv_relu_maxpool_test.onnx")
    print(p1)

    s1 = p1.get_output_shapes()[-1]

    p1_bytes = migraphx.save_buffer(p1)
    assert isinstance(p1_bytes, bytes)

    p2 = migraphx.load_buffer(p1_bytes)
    print(p2)
    s2 = p2.get_output_shapes()[-1]

    assert s1 == s2
    assert p1.sort() == p2.sort()


def create_buffer(t, data, shape):
    a = array.array(t, data)
    if sys.version_info >= (3, 0):
        m = memoryview(a.tobytes())
        return m.cast(t, shape)
    else:
        m = memoryview(a.tostring())
        return m

def test_load_save_arg():
    data = [1,2,3,4]
    buffer1 = create_buffer('f', data, [2,2])
    arg1 = migraphx.argument(buffer1)
    migraphx.argument.save(arg1, 'load_save_arg.msgpack')
    arg2 = migraphx.argument.load('load_save_arg.msgpack')
    assert arg1 == arg2

if __name__ == "__main__":
    test_load_save_arg()
    test_conv_relu('msgpack')
    test_conv_relu('json')
    test_save_load_buffer()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#####################################################################################
from .base import BaseModel, SingleModelDownloadMixin, OptimumHFModelDownloadMixin
from .preprocess import process_image

from transformers import AutoImageProcessor
import timm


class AutoImageProcessorHFMixin(object):

    _processor = None

    @property
    def processor(self):
        if self._processor is None:
            self._processor = AutoImageProcessor.from_pretrained(self.model_id)
        return self._processor

    def preprocess(self, image_data):
        return self.processor(image_data, return_tensors="np")


class ResNet50_v1(SingleModelDownloadMixin, BaseModel):
    @property
    def model_id(self):
        return "https://zenodo.org/record/2592612/files/resnet50_v1.onnx"

    @staticmethod
    def name():
        return "resnet50_v1"

    def preprocess(self, image_data):
        IMAGENET_MEANS = [123.68, 116.78, 103.94]  # RGB
        return {
            "input_tensor:0": process_image(image_data, means=IMAGENET_MEANS)
        }


class ResNet50_v1_5(OptimumHFModelDownloadMixin, AutoImageProcessorHFMixin,
                    BaseModel):
    @property
    def model_id(self):
        return "microsoft/resnet-50"

    @staticmethod
    def name():
        return "resnet50_v1.5"


class VitBasePatch16_224(OptimumHFModelDownloadMixin,
                         AutoImageProcessorHFMixin, BaseModel):
    @property
    def model_id(self):
        return "google/vit-base-patch16-224"

    @staticmethod
    def name():
        return "vit-base-patch16-224"


class TIMM_MobileNetv3_large(OptimumHFModelDownloadMixin, BaseModel):
    def __init__(self):
        data_config = timm.data.resolve_model_data_config(
            timm.create_model(self.model_id, pretrained=True))
        self.processor = timm.data.create_transform(**data_config,
                                                    is_training=False)

    @property
    def model_id(self):
        return "timm/mobilenetv3_large_100.ra_in1k"

    @staticmethod
    def name():
        return "timm-mobilenetv3-large"

    def preprocess(self, image_data):
        return {
            "pixel_values":
            self.processor(image_data).unsqueeze(0).cpu().detach().numpy()
        }


# TODO enable it when BiT is supported by optimum
class Bit50(OptimumHFModelDownloadMixin, AutoImageProcessorHFMixin, BaseModel):
    @property
    def model_id(self):
        return "google/bit-50"

    @staticmethod
    def name():
        return "bit-50"



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#####################################################################################
import onnx
import logging
import os
import requests


def get_model_io(model_path):
    model = onnx.load(model_path)
    outputs = [node.name for node in model.graph.output]

    input_all = [node.name for node in model.graph.input]
    input_initializer = [node.name for node in model.graph.initializer]
    inputs = list(set(input_all) - set(input_initializer))

    return inputs, outputs


def numpy_to_pb(name, np_data, out_filename):
    """Convert numpy data to a protobuf file."""

    tensor = onnx.numpy_helper.from_array(np_data, name)
    onnx.save_tensor(tensor, out_filename)


def download(url, filename, quiet=False):
    try:
        from tqdm import tqdm
    except:
        quiet = True

    logging.debug(f"Download {filename} from {url}")
    response = requests.get(url, stream=True)

    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 1024
    if not quiet:
        progress_bar = tqdm(total=total_size_in_bytes,
                            unit="iB",
                            unit_scale=True)

    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "wb") as f:
        for data in response.iter_content(block_size):
            if not quiet:
                progress_bar.update(len(data))
            f.write(data)
    if not quiet:
        progress_bar.close()


def get_imagenet_classes():
    url = "https://raw.githubusercontent.com/pytorch/hub/master/imagenet_classes.txt"
    response = requests.get(url)
    return response.text.split("\n")



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################

import os
import torch
from transformers import (CLIPTokenizer, T5TokenizerFast, CLIPTextModel,
                          T5EncoderModel)
from diffusers import FluxTransformer2DModel, AutoencoderKL, FlowMatchEulerDiscreteScheduler
import migraphx as mgx


class MGXModel:
    def __init__(self,
                 model,
                 input_shapes=None,
                 fp16=False,
                 bf16=True,
                 exhaustive_tune=False,
                 config=None):

        if isinstance(model, mgx.program):
            self.model = model

        elif isinstance(model, str) and os.path.isfile(model):

            if model.endswith(".mxr"):
                self.model = mgx.load(model, format="msgpack")

            elif model.endswith(".onnx"):
                if not input_shapes:
                    raise ValueError(
                        "input_shapes need to be specified for loading a .onnx file"
                    )
                self.model = mgx.parse_onnx(model, map_input_dims=input_shapes)
                if fp16:
                    mgx.quantize_fp16(self.model)
                elif bf16:
                    mgx.quantize_bf16(self.model)
                self.model.compile(mgx.get_target("gpu"),
                                   exhaustive_tune=exhaustive_tune,
                                   offload_copy=False)
            else:
                raise ValueError(
                    f"File type not recognized (should eend with .mxr or .onnx): {model}"
                )
        else:
            raise ValueError(
                "model should be a migraphx.program object or path to .mxr/.onnx file"
            )
        self.config = config

        self.mgx_to_torch_dtype_dict = {
            "bool_type": torch.bool,
            "uint8_type": torch.uint8,
            "int8_type": torch.int8,
            "int16_type": torch.int16,
            "int32_type": torch.int32,
            "int64_type": torch.int64,
            "float_type": torch.float32,
            "double_type": torch.float64,
            "half_type": torch.float16,
        }
        self.torch_to_mgx_dtype_dict = {
            v: k
            for k, v in self.mgx_to_torch_dtype_dict.items()
        }

        self.input_names = []
        self.output_names = []

        for n in self.model.get_parameter_names():
            if "main:#output_" in n:
                self.output_names.append(n)
            else:
                self.input_names.append(n)

        self.torch_buffers = {}
        self.mgx_args = {}

        self.start_events = []
        self.end_events = []

        self.prealloc_buffers(self.output_names)

    def run_async(self, stream=None, **inputs):
        if stream is None:
            stream = torch.cuda.current_stream()

        for name, tensor in inputs.items():
            self.mgx_args[name] = self.tensor_to_arg(tensor)

        self.start_events.append(torch.cuda.Event(enable_timing=True))
        self.end_events.append(torch.cuda.Event(enable_timing=True))

        self.start_events[-1].record()
        self.model.run_async(self.mgx_args, stream.cuda_stream, "ihipStream_t")
        self.end_events[-1].record()

        return {p: self.torch_buffers[p] for p in self.output_names}

    def save_model(self, path):
        os.makedirs(os.path.dirname(path), exist_ok=True)
        mgx.save(self.model, path, format="msgpack")

    def tensor_to_arg(self, tensor):
        mgx_shape = mgx.shape(type=self.torch_to_mgx_dtype_dict[tensor.dtype],
                              lens=list(tensor.size()),
                              strides=list(tensor.stride()))
        return mgx.argument_from_pointer(mgx_shape, tensor.data_ptr())

    def prealloc_buffers(self, param_names):
        for param_name in param_names:
            param_shape = self.model.get_parameter_shapes()[param_name]

            type_str, lens = param_shape.type_string(), param_shape.lens()
            strides = param_shape.strides()
            torch_dtype = self.mgx_to_torch_dtype_dict[type_str]
            tensor = torch.empty_strided(lens,
                                         strides,
                                         dtype=torch_dtype,
                                         device=torch.cuda.current_device())
            self.torch_buffers[param_name] = tensor
            self.mgx_args[param_name] = self.tensor_to_arg(tensor)

    def get_run_times(self):
        return [
            s.elapsed_time(e)
            for s, e in zip(self.start_events, self.end_events)
        ]

    def clear_events(self):
        self.start_events = []
        self.end_events = []


def get_scheduler(local_dir, hf_model_path, scheduler_dir="scheduler"):
    scheduler_local_dir = os.path.join(local_dir, scheduler_dir)
    scheduler_cls = FlowMatchEulerDiscreteScheduler

    if not os.path.exists(scheduler_local_dir):
        model = scheduler_cls.from_pretrained(hf_model_path,
                                              subfolder=scheduler_dir)
        model.save_pretrained(scheduler_local_dir)
    else:
        print(f"Loading {scheduler_cls} scheduler from {scheduler_local_dir}")
        model = scheduler_cls.from_pretrained(scheduler_local_dir)

    return model


def get_tokenizer(local_dir,
                  hf_model_path,
                  tokenizer_type="clip",
                  tokenizer_dir="tokenizer"):
    tokenizer_local_dir = os.path.join(local_dir, tokenizer_dir)
    if tokenizer_type == "clip":
        tokenizer_class = CLIPTokenizer
    elif tokenizer_type == "t5":
        tokenizer_class = T5TokenizerFast
    else:
        raise ValueError(f"Unsupported tokenizer: {tokenizer_type}")

    if not os.path.exists(tokenizer_local_dir):
        model = tokenizer_class.from_pretrained(hf_model_path,
                                                subfolder=tokenizer_dir)
        model.save_pretrained(tokenizer_local_dir)
    else:
        print(f"Loading {tokenizer_type} tokenizer from {tokenizer_local_dir}")
        model = tokenizer_class.from_pretrained(tokenizer_local_dir)

    return model


def get_local_path(local_dir, model_dir):
    model_local_dir = os.path.join(local_dir, model_dir)
    if not os.path.exists(model_local_dir):
        os.makedirs(model_local_dir)
    return model_local_dir


def get_clip_model(local_dir,
                   hf_model_path,
                   compiled_dir,
                   model_dir="text_encoder",
                   torch_dtype=torch.float32,
                   bs=1,
                   exhaustive_tune=False,
                   fp16=False,
                   bf16=True):
    clip_local_dir = get_local_path(local_dir, model_dir)
    onnx_file = "model.onnx"
    onnx_path = os.path.join(clip_local_dir, onnx_file)

    def get_compiled_file_name():
        name = f"model_b{bs}"
        if fp16:
            name += "_fp16"
        elif bf16:
            name += "_bf16"

        if exhaustive_tune: name += "_exh"
        return name + ".mxr"

    clip_compiled_dir = get_local_path(compiled_dir, model_dir)
    mxr_file = get_compiled_file_name()
    mxr_path = os.path.join(clip_compiled_dir, mxr_file)

    if os.path.isfile(mxr_path):
        print(f"found compiled model.. loading CLIP encoder from {mxr_path}")
        model = MGXModel(mxr_path)
        return model

    sample_inputs = (torch.zeros(bs, 77, dtype=torch.int32), )
    input_names = ["input_ids"]
    if not os.path.isfile(onnx_path):
        print("ONNX file not found.. exporting CLIP encoder to ONNX")
        model = CLIPTextModel.from_pretrained(hf_model_path,
                                              subfolder=model_dir,
                                              torch_dtype=torch_dtype)

        output_names = ["text_embeddings"]
        dynamic_axes = {"input_ids": {0: 'B'}, "text_embeddings": {0: 'B'}}

        # CLIP export requires nightly pytorch due to bug in onnx parser
        with torch.inference_mode():
            torch.onnx.export(model,
                              sample_inputs,
                              onnx_path,
                              export_params=True,
                              input_names=input_names,
                              output_names=output_names,
                              dynamic_axes=dynamic_axes)

    assert os.path.isfile(onnx_path)
    print(f"Generating MXR from ONNX file: {onnx_path}")
    input_shapes = {
        n: list(t.size())
        for n, t in zip(input_names, sample_inputs)
    }
    model = MGXModel(onnx_path,
                     input_shapes=input_shapes,
                     exhaustive_tune=exhaustive_tune,
                     fp16=fp16,
                     bf16=bf16)
    model.save_model(os.path.join(clip_compiled_dir, get_compiled_file_name()))

    return model

    # migraphx-driver perf FLUX.1-schnell/text_encoder/model.onnx --input-dim @input_ids 1 77 --fill1 input_ids --fp16


def get_t5_model(local_dir,
                 hf_model_path,
                 compiled_dir,
                 max_len=512,
                 model_dir="text_encoder_2",
                 torch_dtype=torch.float32,
                 bs=1,
                 exhaustive_tune=False,
                 fp16=False,
                 bf16=True):
    t5_local_dir = get_local_path(local_dir, model_dir)
    onnx_file = "model.onnx"
    onnx_path = os.path.join(t5_local_dir, onnx_file)

    def get_compiled_file_name():
        name = f"model_b{bs}"
        name += f"_l{max_len}"
        if fp16:
            name += "_fp16"
        elif bf16:
            name += "_bf16"
        if exhaustive_tune: name += "_exh"
        return name + ".mxr"

    t5_compiled_dir = get_local_path(compiled_dir, model_dir)
    mxr_file = get_compiled_file_name()
    mxr_path = os.path.join(t5_compiled_dir, mxr_file)

    if os.path.isfile(mxr_path):
        print(f"found compiled model.. loading T5 encoder from {mxr_path}")
        model = MGXModel(mxr_path)
        return model

    sample_inputs = (torch.zeros(bs, max_len, dtype=torch.int32), )
    input_names = ["input_ids"]
    if not os.path.isfile(onnx_path):
        model = T5EncoderModel.from_pretrained(hf_model_path,
                                               subfolder=model_dir,
                                               torch_dtype=torch_dtype)
        output_names = ["text_embeddings"]
        dynamic_axes = {"input_ids": {0: 'B'}, "text_embeddings": {0: 'B'}}

        with torch.inference_mode():
            torch.onnx.export(model,
                              sample_inputs,
                              onnx_path,
                              export_params=True,
                              input_names=input_names,
                              output_names=output_names,
                              dynamic_axes=dynamic_axes)

    assert os.path.isfile(onnx_path)
    print(f"Generating MXR from ONNX file: {onnx_path}")
    input_shapes = {
        n: list(t.size())
        for n, t in zip(input_names, sample_inputs)
    }
    model = MGXModel(onnx_path,
                     input_shapes=input_shapes,
                     exhaustive_tune=exhaustive_tune,
                     fp16=fp16,
                     bf16=bf16)
    model.save_model(os.path.join(t5_compiled_dir, get_compiled_file_name()))

    return model

    # migraphx-driver perf FLUX.1-schnell/text_encoder_2/model.onnx --input-dim @input_ids 1 512 --fill1 input_ids --fp16


## Following decorators required to apply fp16 inference patch to the transformer blocks
## Note that we do not export fp16 weights directly to ONNX to allow migraphx to
##  perform optimizations before quantizing down to fp16. This results in better
##  accuracy compared to exporting fp16 directly to onnx
def transformer_block_clip_wrapper(fn):
    def new_forward(*args, **kwargs):
        encoder_hidden_states, hidden_states = fn(*args, **kwargs)
        return encoder_hidden_states.clip(-65504, 65504), hidden_states

    return new_forward


def single_transformer_block_clip_wrapper(fn):
    def new_forward(*args, **kwargs):
        hidden_states = fn(*args, **kwargs)
        return hidden_states.clip(-65504, 65504)

    return new_forward


def add_output_clippings_for_fp16(model):
    for b in model.transformer_blocks:
        b.forward = transformer_block_clip_wrapper(b.forward)

    for b in model.single_transformer_blocks:
        b.forward = single_transformer_block_clip_wrapper(b.forward)


def get_flux_transformer_model(local_dir,
                               hf_model_path,
                               compiled_dir,
                               img_height=1024,
                               img_width=1024,
                               compression_factor=8,
                               max_len=512,
                               model_dir="transformer",
                               torch_dtype=torch.float32,
                               bs=1,
                               exhaustive_tune=False,
                               fp16=False,
                               bf16=True):

    transformer_local_dir = get_local_path(local_dir, model_dir)
    onnx_file = "model_clip_workaround.onnx" if fp16 else "model.onnx"
    onnx_path = os.path.join(transformer_local_dir, onnx_file)
    latent_h, latent_w = img_height // compression_factor, img_width // compression_factor

    def get_compiled_file_name():
        name = f"model_b{bs}"
        name += f"_h{latent_h}_w{latent_w}_l{max_len}"
        if fp16:
            name += "_fp16"
        elif bf16:
            name += "_bf16"
        if exhaustive_tune: name += "_exh"
        return name + ".mxr"

    transformer_compiled_dir = get_local_path(compiled_dir, model_dir)
    mxr_file = get_compiled_file_name()
    mxr_path = os.path.join(transformer_compiled_dir, mxr_file)

    config = FluxTransformer2DModel.load_config(hf_model_path,
                                                subfolder=model_dir)

    if os.path.isfile(mxr_path):
        print(
            f"found compiled model.. loading flux transformer from {mxr_path}")
        model = MGXModel(mxr_path, config=config)
        return model

    sample_inputs = (
        torch.randn(bs, (latent_h // 2) * (latent_w // 2),
                    config["in_channels"],
                    dtype=torch_dtype),
        torch.randn(bs,
                    max_len,
                    config['joint_attention_dim'],
                    dtype=torch_dtype),
        torch.randn(bs, config['pooled_projection_dim'], dtype=torch_dtype),
        torch.tensor([1.] * bs, dtype=torch_dtype),
        torch.randn((latent_h // 2) * (latent_w // 2), 3, dtype=torch_dtype),
        torch.randn(max_len, 3, dtype=torch_dtype),
        torch.tensor([1.] * bs, dtype=torch_dtype),
    )

    input_names = [
        'hidden_states', 'encoder_hidden_states', 'pooled_projections',
        'timestep', 'img_ids', 'txt_ids', 'guidance'
    ]
    if not os.path.isfile(onnx_path):
        model = FluxTransformer2DModel.from_pretrained(hf_model_path,
                                                       subfolder=model_dir,
                                                       torch_dtype=torch_dtype)

        if fp16:
            print("applying fp16 clip workarounds to transformer")
            add_output_clippings_for_fp16(model)

        output_names = ["latent"]
        dynamic_axes = {
            'hidden_states': {
                0: 'B',
                1: 'latent_dim'
            },
            'encoder_hidden_states': {
                0: 'B',
                1: 'L'
            },
            'pooled_projections': {
                0: 'B'
            },
            'timestep': {
                0: 'B'
            },
            'img_ids': {
                0: 'latent_dim'
            },
            'txt_ids': {
                0: 'L'
            },
            'guidance': {
                0: 'B'
            },
        }

        with torch.inference_mode():
            torch.onnx.export(model,
                              sample_inputs,
                              onnx_path,
                              export_params=True,
                              input_names=input_names,
                              output_names=output_names,
                              dynamic_axes=dynamic_axes)

    assert os.path.isfile(onnx_path)
    print(f"Generating MXR from ONNX file: {onnx_path}")
    input_shapes = {
        n: list(t.size())
        for n, t in zip(input_names, sample_inputs)
    }
    model = MGXModel(onnx_path,
                     input_shapes=input_shapes,
                     exhaustive_tune=exhaustive_tune,
                     fp16=fp16,
                     bf16=bf16,
                     config=config)
    model.save_model(
        os.path.join(transformer_compiled_dir, get_compiled_file_name()))

    return model
    # migraphx-driver perf FLUX.1-schnell/transformer/model.onnx --input-dim @hidden_states 1 4096 64 @encoder_hidden_states 1 512 4096 @pooled_projections 1 768 @timestep 1 @img_ids 4096 3 @txt_ids 512 3 --fp16
    # migraphx-driver perf FLUX.1-dev/transformer/model.onnx --input-dim @hidden_states 1 4096 64 @encoder_hidden_states 1 512 4096 @pooled_projections 1 768 @timestep 1 @img_ids 4096 3 @txt_ids 512 3 @guidance 1 --fp16


def get_vae_model(local_dir,
                  hf_model_path,
                  compiled_dir,
                  img_height=1024,
                  img_width=1024,
                  compression_factor=8,
                  model_dir="vae",
                  torch_dtype=torch.float32,
                  bs=1,
                  exhaustive_tune=False,
                  fp16=False,
                  bf16=True):

    vae_local_dir = get_local_path(local_dir, model_dir)
    onnx_file = "model.onnx"
    onnx_path = os.path.join(vae_local_dir, onnx_file)
    latent_h, latent_w = img_height // compression_factor, img_width // compression_factor

    def get_compiled_file_name():
        name = f"model_b{bs}"
        name += f"_h{latent_h}_w{latent_w}"
        if fp16:
            name += "_fp16"
        elif bf16:
            name += "_bf16"
        if exhaustive_tune: name += "_exh"
        return name + ".mxr"

    vae_compiled_dir = get_local_path(compiled_dir, model_dir)
    mxr_file = get_compiled_file_name()
    mxr_path = os.path.join(vae_compiled_dir, mxr_file)

    config = AutoencoderKL.load_config(hf_model_path, subfolder=model_dir)

    if os.path.isfile(mxr_path):
        print(f"found compiled model.. loading VAE decoder from {mxr_path}")
        model = MGXModel(mxr_path, config=config)
        return model

    sample_inputs = (torch.randn(bs,
                                 config['latent_channels'],
                                 latent_h,
                                 latent_w,
                                 dtype=torch_dtype), )
    input_names = ["latent"]
    if not os.path.isfile(onnx_path):
        model = AutoencoderKL.from_pretrained(hf_model_path,
                                              subfolder=model_dir,
                                              torch_dtype=torch_dtype)
        model.forward = model.decode

        output_names = ["images"]
        dynamic_axes = {
            'latent': {
                0: 'B',
                2: 'H',
                3: 'W'
            },
            'images': {
                0: 'B',
                2: '8H',
                3: '8W'
            }
        }

        with torch.inference_mode():
            torch.onnx.export(model,
                              sample_inputs,
                              onnx_path,
                              export_params=True,
                              input_names=input_names,
                              output_names=output_names,
                              dynamic_axes=dynamic_axes)

    assert os.path.isfile(onnx_path)
    print(f"Generating MXR from ONNX file: {onnx_path}")
    input_shapes = {
        n: list(t.size())
        for n, t in zip(input_names, sample_inputs)
    }
    model = MGXModel(onnx_path,
                     input_shapes=input_shapes,
                     exhaustive_tune=exhaustive_tune,
                     fp16=fp16,
                     bf16=bf16,
                     config=config)
    model.save_model(os.path.join(vae_compiled_dir, get_compiled_file_name()))

    return model
    # migraphx-driver perf FLUX.1-schnell/vae/model.onnx --input-dim @latent 1 16 128 128 --fp16



#####################################################################################
#  The MIT License (MIT)
#
#  Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
#####################################################################################

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import re

html_theme = "rocm_docs_theme"
html_theme_options = {"flavor": "rocm-docs-home"}

templates_path = ["."]  # Use the current folder for templates

setting_all_article_info = True
all_article_info_os = ["linux"]

with open('../CMakeLists.txt', encoding='utf-8') as f:
    match = re.search(r'.*\brocm_setup_version\(VERSION\s+([0-9.]+)[^0-9.]+',
                      f.read())
    if not match:
        raise ValueError("VERSION not found!")
    version_number = match[1]

# for PDF output on Read the Docs
project = "MIGraphX"
author = "Advanced Micro Devices, Inc."
copyright = "Copyright (c) 2024 Advanced Micro Devices, Inc. All rights reserved."
version = version_number
release = version_number

extensions = ["rocm_docs", "rocm_docs.doxygen", "sphinx_collapse"]
external_toc_path = "./sphinx/_toc.yml"
doxygen_root = "doxygen"
doxysphinx_enabled = False
doxygen_project = {
    "name": "doxygen",
    "path": "doxygen/xml",
}

html_title = f"{project} {version_number} documentation"

external_projects_current_project = "amdmigraphx"



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
# All pre- and post-processing methods used below are borrowed from the ONNX MOdel Zoo
# https://github.com/onnx/models/tree/master/vision/object_detection_segmentation/yolov4

import numpy as np
import cv2
from scipy import special
import colorsys
import random


# this function is from tensorflow-yolov4-tflite/core/utils.py
def image_preprocess(image, target_size, gt_boxes=None):

    ih, iw = target_size
    h, w, _ = image.shape

    scale = min(iw / w, ih / h)
    nw, nh = int(scale * w), int(scale * h)
    image_resized = cv2.resize(image, (nw, nh))

    image_padded = np.full(shape=[ih, iw, 3], fill_value=128.0)
    dw, dh = (iw - nw) // 2, (ih - nh) // 2
    image_padded[dh:nh + dh, dw:nw + dw, :] = image_resized
    image_padded = image_padded / 255.

    if gt_boxes is None:
        return image_padded

    else:
        gt_boxes[:, [0, 2]] = gt_boxes[:, [0, 2]] * scale + dw
        gt_boxes[:, [1, 3]] = gt_boxes[:, [1, 3]] * scale + dh
        return image_padded, gt_boxes


def get_anchors(anchors_path, tiny=False):
    '''loads the anchors from a file'''
    with open(anchors_path) as f:
        anchors = f.readline()
    anchors = np.array(anchors.split(','), dtype=np.float32)
    return anchors.reshape(3, 3, 2)


def postprocess_bbbox(pred_bbox, ANCHORS, STRIDES, XYSCALE=[1, 1, 1]):
    '''define anchor boxes'''
    for i, pred in enumerate(pred_bbox):
        conv_shape = pred.shape
        output_size = conv_shape[1]
        conv_raw_dxdy = pred[:, :, :, :, 0:2]
        conv_raw_dwdh = pred[:, :, :, :, 2:4]
        xy_grid = np.meshgrid(np.arange(output_size), np.arange(output_size))
        xy_grid = np.expand_dims(np.stack(xy_grid, axis=-1), axis=2)

        xy_grid = np.tile(np.expand_dims(xy_grid, axis=0), [1, 1, 1, 3, 1])
        xy_grid = xy_grid.astype(np.float)

        pred_xy = ((special.expit(conv_raw_dxdy) * XYSCALE[i]) - 0.5 *
                   (XYSCALE[i] - 1) + xy_grid) * STRIDES[i]
        pred_wh = (np.exp(conv_raw_dwdh) * ANCHORS[i])
        pred[:, :, :, :, 0:4] = np.concatenate([pred_xy, pred_wh], axis=-1)

    pred_bbox = [np.reshape(x, (-1, np.shape(x)[-1])) for x in pred_bbox]
    pred_bbox = np.concatenate(pred_bbox, axis=0)
    return pred_bbox


def postprocess_boxes(pred_bbox, org_img_shape, input_size, score_threshold):
    '''remove boundary boxs with a low detection probability'''
    valid_scale = [0, np.inf]
    pred_bbox = np.array(pred_bbox)

    pred_xywh = pred_bbox[:, 0:4]
    pred_conf = pred_bbox[:, 4]
    pred_prob = pred_bbox[:, 5:]

    # (1) (x, y, w, h) --> (xmin, ymin, xmax, ymax)
    pred_coor = np.concatenate([
        pred_xywh[:, :2] - pred_xywh[:, 2:] * 0.5,
        pred_xywh[:, :2] + pred_xywh[:, 2:] * 0.5
    ],
                               axis=-1)
    # (2) (xmin, ymin, xmax, ymax) -> (xmin_org, ymin_org, xmax_org, ymax_org)
    org_h, org_w = org_img_shape
    resize_ratio = min(input_size / org_w, input_size / org_h)

    dw = (input_size - resize_ratio * org_w) / 2
    dh = (input_size - resize_ratio * org_h) / 2

    pred_coor[:, 0::2] = 1.0 * (pred_coor[:, 0::2] - dw) / resize_ratio
    pred_coor[:, 1::2] = 1.0 * (pred_coor[:, 1::2] - dh) / resize_ratio

    # (3) clip some boxes that are out of range
    pred_coor = np.concatenate([
        np.maximum(pred_coor[:, :2], [0, 0]),
        np.minimum(pred_coor[:, 2:], [org_w - 1, org_h - 1])
    ],
                               axis=-1)
    invalid_mask = np.logical_or((pred_coor[:, 0] > pred_coor[:, 2]),
                                 (pred_coor[:, 1] > pred_coor[:, 3]))
    pred_coor[invalid_mask] = 0

    # (4) discard some invalid boxes
    bboxes_scale = np.sqrt(
        np.multiply.reduce(pred_coor[:, 2:4] - pred_coor[:, 0:2], axis=-1))
    scale_mask = np.logical_and((valid_scale[0] < bboxes_scale),
                                (bboxes_scale < valid_scale[1]))

    # (5) discard some boxes with low scores
    classes = np.argmax(pred_prob, axis=-1)
    scores = pred_conf * pred_prob[np.arange(len(pred_coor)), classes]
    score_mask = scores > score_threshold
    mask = np.logical_and(scale_mask, score_mask)
    coors, scores, classes = pred_coor[mask], scores[mask], classes[mask]

    return np.concatenate(
        [coors, scores[:, np.newaxis], classes[:, np.newaxis]], axis=-1)


def bboxes_iou(boxes1, boxes2):
    '''calculate the Intersection Over Union value'''
    boxes1 = np.array(boxes1)
    boxes2 = np.array(boxes2)

    boxes1_area = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] -
                                                       boxes1[..., 1])
    boxes2_area = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] -
                                                       boxes2[..., 1])

    left_up = np.maximum(boxes1[..., :2], boxes2[..., :2])
    right_down = np.minimum(boxes1[..., 2:], boxes2[..., 2:])

    inter_section = np.maximum(right_down - left_up, 0.0)
    inter_area = inter_section[..., 0] * inter_section[..., 1]
    union_area = boxes1_area + boxes2_area - inter_area
    ious = np.maximum(1.0 * inter_area / union_area, np.finfo(np.float32).eps)

    return ious


def nms(bboxes, iou_threshold, sigma=0.3, method='nms'):
    """
    :param bboxes: (xmin, ymin, xmax, ymax, score, class)

    Note: soft-nms, https://arxiv.org/pdf/1704.04503.pdf
          https://github.com/bharatsingh430/soft-nms
    """
    classes_in_img = list(set(bboxes[:, 5]))
    best_bboxes = []

    for cls in classes_in_img:
        cls_mask = (bboxes[:, 5] == cls)
        cls_bboxes = bboxes[cls_mask]

        while len(cls_bboxes) > 0:
            max_ind = np.argmax(cls_bboxes[:, 4])
            best_bbox = cls_bboxes[max_ind]
            best_bboxes.append(best_bbox)
            cls_bboxes = np.concatenate(
                [cls_bboxes[:max_ind], cls_bboxes[max_ind + 1:]])
            iou = bboxes_iou(best_bbox[np.newaxis, :4], cls_bboxes[:, :4])
            weight = np.ones((len(iou), ), dtype=np.float32)

            assert method in ['nms', 'soft-nms']

            if method == 'nms':
                iou_mask = iou > iou_threshold
                weight[iou_mask] = 0.0

            if method == 'soft-nms':
                weight = np.exp(-(1.0 * iou**2 / sigma))

            cls_bboxes[:, 4] = cls_bboxes[:, 4] * weight
            score_mask = cls_bboxes[:, 4] > 0.
            cls_bboxes = cls_bboxes[score_mask]

    return best_bboxes


def read_class_names(class_file_name):
    '''loads class name from a file'''
    names = {}
    with open(class_file_name, 'r') as data:
        for ID, name in enumerate(data):
            names[ID] = name.strip('\n')
    return names


def draw_bbox(image,
              bboxes,
              classes=read_class_names("./utilities/coco.names"),
              show_label=True):
    """
    bboxes: [x_min, y_min, x_max, y_max, probability, cls_id] format coordinates.
    """

    num_classes = len(classes)
    image_h, image_w, _ = image.shape
    hsv_tuples = [(1.0 * x / num_classes, 1., 1.) for x in range(num_classes)]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(
        map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)),
            colors))

    random.seed(0)
    random.shuffle(colors)
    random.seed(None)

    for i, bbox in enumerate(bboxes):
        coor = np.array(bbox[:4], dtype=np.int32)
        fontScale = 0.5
        score = bbox[4]
        class_ind = int(bbox[5])
        bbox_color = colors[class_ind]
        bbox_thick = int(0.6 * (image_h + image_w) / 600)
        c1, c2 = (coor[0], coor[1]), (coor[2], coor[3])
        cv2.rectangle(image, c1, c2, bbox_color, bbox_thick)

        if show_label:
            bbox_mess = '%s: %.2f' % (classes[class_ind], score)
            t_size = cv2.getTextSize(bbox_mess,
                                     0,
                                     fontScale,
                                     thickness=bbox_thick // 2)[0]
            cv2.rectangle(image, c1,
                          (c1[0] + t_size[0], c1[1] - t_size[1] - 3),
                          bbox_color, -1)

            cv2.putText(image,
                        bbox_mess, (c1[0], c1[1] - 2),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale, (0, 0, 0),
                        bbox_thick // 2,
                        lineType=cv2.LINE_AA)

    return image



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#####################################################################################
from .base import BaseModel, OptimumHFModelDownloadMixin
from ..utils import get_imagenet_classes
from transformers import AutoProcessor


class AutoProcessorHFMixin(object):

    _processor = None

    @property
    def processor(self):
        if self._processor is None:
            self._processor = AutoProcessor.from_pretrained(self.model_id)
        return self._processor

    def preprocess(self, images, text, max_length=384):
        return self.processor(images=images,
                              text=text,
                              max_length=max_length,
                              padding='max_length',
                              return_tensors="np")


class ClipVitLargePatch14(OptimumHFModelDownloadMixin, AutoProcessorHFMixin,
                          BaseModel):
    def __init__(self):
        import random
        random.seed(42)
        # cache a few labels, the full 1000 label is overkill
        self.imagenet_labels = random.sample(get_imagenet_classes(), 10)

    @property
    def model_id(self):
        return "openai/clip-vit-large-patch14"

    @staticmethod
    def name():
        return "clip-vit-large-patch14"

    def preprocess(self, *args, **kwargs):
        # extend image with imagenet labels
        new_args, new_kwargs = list(args), kwargs
        new_args.append(self.imagenet_labels)
        new_kwargs["max_length"] = 77
        result = super().preprocess(*new_args, **new_kwargs)
        return result



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#####################################################################################
import os
import onnxruntime as ort

from ..utils import get_model_io, numpy_to_pb


def _inference(session, input_data_map, outputs, folder_name):
    input_pb_name = "input_{}.pb"
    output_pb_name = "output_{}.pb"
    os.makedirs(folder_name, exist_ok=True)
    for input_idx, (input_name,
                    input_data) in enumerate(input_data_map.items()):
        numpy_to_pb(input_name, input_data,
                    f"{folder_name}/{input_pb_name.format(input_idx)}")

    ort_result = session.run(outputs, input_data_map)
    output_data_map = {
        output_name: result_data
        for (output_name, result_data) in zip(outputs, ort_result)
    }
    for output_idx, (output_name,
                     result_data) in enumerate(output_data_map.items()):
        numpy_to_pb(output_name, result_data,
                    f"{folder_name}/{output_pb_name.format(output_idx)}")

    return output_data_map


def generate_test_dataset(model,
                          dataset,
                          output_folder_prefix=None,
                          sample_limit=None,
                          decode_limit=None):
    assert model.is_diffuser() == False, "Only base and decoder supported"
    output_path = f"{output_folder_prefix or 'generated'}/{dataset.name()}/{model.name()}"
    folder_name_prefix = f"{output_path}/test_data_set"

    model_path = model.get_model(output_path)
    inputs, outputs = get_model_io(model_path)

    sess = ort.InferenceSession(model_path)
    print(f"Creating {folder_name_prefix}s...")
    test_idx = 0
    for idx, data in enumerate(dataset):
        input_data_map = dataset.transform(inputs, data, model.preprocess)
        is_eos, decode_idx = False, 0
        while not is_eos and not (decode_limit and decode_limit <= decode_idx):
            folder_name = f"{folder_name_prefix}_{test_idx}"

            output_data_map = _inference(sess, input_data_map, outputs,
                                         folder_name)

            is_eos = not model.is_decoder() or model.decode_step(
                input_data_map, output_data_map)
            test_idx += 1
            decode_idx += 1

        if sample_limit and sample_limit - 1 <= idx:
            break



#  The MIT License (MIT)
#
#  Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#  Copyright (c) 2024 Stability AI
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the 'Software'), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from argparse import ArgumentParser
from diffusers import FlowMatchEulerDiscreteScheduler

from other_impls import SD3Tokenizer

from PIL import Image

import migraphx as mgx
import os
import sys
import torch
import time
from functools import wraps

from hip import hip
from collections import namedtuple
HipEventPair = namedtuple('HipEventPair', ['start', 'end'])


# measurement helper
def measure(fn):
    @wraps(fn)
    def measure_ms(*args, **kwargs):
        start_time = time.perf_counter_ns()
        result = fn(*args, **kwargs)
        end_time = time.perf_counter_ns()
        print(
            f"Elapsed time for {fn.__name__}: {(end_time - start_time) * 1e-6:.4f} ms\n"
        )
        return result

    return measure_ms


def get_args():
    parser = ArgumentParser()
    # Model compile
    parser.add_argument(
        "--onnx-model-path",
        type=str,
        default="models/sd3",
        help="Path to onnx model files.",
    )

    parser.add_argument(
        "--compiled-model-path",
        type=str,
        default=None,
        help=
        "Path to compiled mxr model files. If not set, it will be saved next to the onnx model.",
    )

    parser.add_argument(
        "--fp16",
        choices=["all", "vae", "clip", "mmdit"],
        nargs="+",
        help="Quantize models with fp16 precision.",
    )

    parser.add_argument(
        "--force-compile",
        action="store_true",
        default=False,
        help="Ignore existing .mxr files and override them",
    )

    parser.add_argument(
        "--exhaustive-tune",
        action="store_true",
        default=False,
        help="Perform exhaustive tuning when compiling onnx models",
    )

    parser.add_argument(
        "--skip-t5",
        action="store_true",
        default=False,
        help=
        "Skip the third text encoder. Small accuracy penalty but large memory savings."
    )

    # Runtime
    parser.add_argument(
        "-s",
        "--seed",
        type=int,
        default=42,
        help="Random seed",
    )

    parser.add_argument(
        "-t",
        "--steps",
        type=int,
        default=50,
        help="Number of steps",
    )

    parser.add_argument("-b",
                        "--batch",
                        type=int,
                        default=1,
                        help="Batch count or number of images to produce")

    parser.add_argument(
        "-p",
        "--prompt",
        type=str,
        required=True,
        help="Prompt",
    )

    parser.add_argument(
        "-n",
        "--negative-prompt",
        type=str,
        default="",
        help="Negative prompt",
    )

    parser.add_argument(
        "--scale",
        type=float,
        default=5.0,
        help="Guidance scale",
    )

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="Output name",
    )
    return parser.parse_args()


mgx_to_torch_dtype_dict = {
    "bool_type": torch.bool,
    "uint8_type": torch.uint8,
    "int8_type": torch.int8,
    "int16_type": torch.int16,
    "int32_type": torch.int32,
    "int64_type": torch.int64,
    "float_type": torch.float32,
    "double_type": torch.float64,
    "half_type": torch.float16,
}

torch_to_mgx_dtype_dict = {
    value: key
    for (key, value) in mgx_to_torch_dtype_dict.items()
}


def tensor_to_arg(tensor):
    return mgx.argument_from_pointer(
        mgx.shape(
            **{
                "type": torch_to_mgx_dtype_dict[tensor.dtype],
                "lens": list(tensor.size()),
                "strides": list(tensor.stride())
            }), tensor.data_ptr())


def tensors_to_args(tensors):
    return {name: tensor_to_arg(tensor) for name, tensor in tensors.items()}


def get_output_name(idx):
    return f"main:#output_{idx}"


def copy_tensor_sync(tensor, data):
    tensor.copy_(data)
    torch.cuda.synchronize()


def run_model_sync(model, args):
    model.run(args)
    mgx.gpu_sync()


def allocate_torch_tensors(model):
    input_shapes = model.get_parameter_shapes()
    data_mapping = {
        name: torch.zeros(shape.lens()).to(
            mgx_to_torch_dtype_dict[shape.type_string()]).to(device="cuda")
        for name, shape in input_shapes.items()
    }
    return data_mapping


class StableDiffusionMGX():
    def __init__(self,
                 onnx_model_path,
                 compiled_model_path,
                 fp16,
                 batch,
                 force_compile=False,
                 exhaustive_tune=False,
                 skip_t5=False):

        self.scheduler = FlowMatchEulerDiscreteScheduler.from_pretrained(
            "stabilityai/stable-diffusion-3-medium-diffusers",
            subfolder="scheduler")

        self.tokenizer = SD3Tokenizer()
        self.device = "cuda"
        self.skip_t5 = skip_t5

        if fp16 is None:
            fp16 = []
        elif "all" in fp16:
            fp16 = ["vae", "clip", "mmdit"]

        self.batch = batch

        print("Load models...")
        self.models = {
            "vae":
            StableDiffusionMGX.load_mgx_model(
                "vae_decoder", {"latent_sample": [self.batch, 16, 128, 128]},
                onnx_model_path,
                compiled_model_path=compiled_model_path,
                use_fp16="vae" in fp16,
                force_compile=force_compile,
                exhaustive_tune=exhaustive_tune,
                offload_copy=False,
                batch=self.batch),
            "clip-l":
            StableDiffusionMGX.load_mgx_model(
                "text_encoder", {"input_ids": [1, 77]},
                onnx_model_path,
                compiled_model_path=compiled_model_path,
                use_fp16="clip" in fp16,
                force_compile=force_compile,
                exhaustive_tune=exhaustive_tune,
                offload_copy=False),
            "clip-g":
            StableDiffusionMGX.load_mgx_model(
                "text_encoder_2", {"input_ids": [1, 77]},
                onnx_model_path,
                compiled_model_path=compiled_model_path,
                use_fp16="clip" in fp16,
                force_compile=force_compile,
                exhaustive_tune=exhaustive_tune,
                offload_copy=False),
            "mmdit":
            StableDiffusionMGX.load_mgx_model(
                "transformer", {
                    "hidden_states": [2 * self.batch, 16, 128, 128],
                    "timestep": [2 * self.batch],
                    "encoder_hidden_states": [2 * self.batch, 154, 4096],
                    "pooled_projections": [2 * self.batch, 2048],
                },
                onnx_model_path,
                compiled_model_path=compiled_model_path,
                use_fp16="mmdit" in fp16,
                force_compile=force_compile,
                exhaustive_tune=exhaustive_tune,
                offload_copy=False,
                batch=self.batch)
        }

        self.tensors = {
            "clip-g": allocate_torch_tensors(self.models["clip-g"]),
            "clip-l": allocate_torch_tensors(self.models["clip-l"]),
            # "t5xxl": allocate_torch_tensors(self.models["t5xxl"]),
            "mmdit": allocate_torch_tensors(self.models["mmdit"]),
            "vae": allocate_torch_tensors(self.models["vae"]),
        }

        self.model_args = {
            "clip-g": tensors_to_args(self.tensors['clip-g']),
            "clip-l": tensors_to_args(self.tensors['clip-l']),
            # "t5xxl": tensors_to_args(self.tensors['t5xxl']),
            "mmdit": tensors_to_args(self.tensors['mmdit']),
            "vae": tensors_to_args(self.tensors['vae']),
        }

        if not self.skip_t5:
            self.models["t5xxl"] = StableDiffusionMGX.load_mgx_model(
                "text_encoder_3", {"input_ids": [1, 77]},
                onnx_model_path,
                compiled_model_path=compiled_model_path,
                use_fp16="clip" in fp16,
                force_compile=force_compile,
                exhaustive_tune=exhaustive_tune,
                offload_copy=False)
            self.tensors["t5xxl"] = allocate_torch_tensors(
                self.models["t5xxl"])
            self.model_args["t5xxl"] = tensors_to_args(self.tensors['t5xxl'])

        self.events = {
            "warmup":
            HipEventPair(start=hip.hipEventCreate()[1],
                         end=hip.hipEventCreate()[1]),
            "run":
            HipEventPair(start=hip.hipEventCreate()[1],
                         end=hip.hipEventCreate()[1]),
            "clip":
            HipEventPair(start=hip.hipEventCreate()[1],
                         end=hip.hipEventCreate()[1]),
            "denoise":
            HipEventPair(start=hip.hipEventCreate()[1],
                         end=hip.hipEventCreate()[1]),
            "decode":
            HipEventPair(start=hip.hipEventCreate()[1],
                         end=hip.hipEventCreate()[1]),
        }

        self.stream = hip.hipStreamCreate()[1]

    def cleanup(self):
        for event in self.events.values():
            hip.hipEventDestroy(event.start)
            hip.hipEventDestroy(event.end)
        hip.hipStreamDestroy(self.stream)

    def profile_start(self, name):
        if name in self.events:
            hip.hipEventRecord(self.events[name].start, None)

    def profile_end(self, name):
        if name in self.events:
            hip.hipEventRecord(self.events[name].end, None)

    @measure
    @torch.no_grad()
    def run(self, prompt, negative_prompt, steps, seed, scale):
        torch.cuda.synchronize()
        self.profile_start("run")

        print("Tokenizing prompts...")
        prompt_tokens = self.tokenize(prompt)
        neg_prompt_tokens = self.tokenize(negative_prompt)

        print("Creating text embeddings...")
        self.profile_start("clip")
        prompt_embeddings = self.get_embeddings(prompt_tokens)
        neg_prompt_embeddings = self.get_embeddings(neg_prompt_tokens)
        self.profile_end("clip")

        # fix height and width for now
        # TODO: check for valid height/width combinations
        # and make them member variables
        height = 1024
        width = 1024
        latent = torch.empty(1, 16, height // 8, width // 8, device="cpu")

        generator = torch.manual_seed(seed)
        latent = torch.randn(latent.size(),
                             dtype=torch.float32,
                             layout=latent.layout,
                             generator=generator).to(latent.dtype)

        self.scheduler.set_timesteps(steps)
        timesteps = self.scheduler.timesteps

        print("Running denoising loop...")
        self.profile_start("denoise")
        for step in timesteps:
            latent = self.denoise(latent, prompt_embeddings,
                                  neg_prompt_embeddings, step, scale)

        self.profile_end("denoise")

        latent = (latent / 1.5305) + 0.0609

        self.profile_start("decode")
        print("Decode denoised result...")
        image = self.decode(latent)
        self.profile_end("decode")

        torch.cuda.synchronize()
        self.profile_end("run")
        return image

    def print_summary(self, denoise_steps):
        print('WARMUP\t{:>9.2f} ms'.format(
            hip.hipEventElapsedTime(self.events['warmup'].start,
                                    self.events['warmup'].end)[1]))
        print('CLIP\t{:>9.2f} ms'.format(
            hip.hipEventElapsedTime(self.events['clip'].start,
                                    self.events['clip'].end)[1]))
        print('mmditx{}\t{:>9.2f} ms'.format(
            str(denoise_steps),
            hip.hipEventElapsedTime(self.events['denoise'].start,
                                    self.events['denoise'].end)[1]))
        print('VAE-Dec\t{:>9.2f} ms'.format(
            hip.hipEventElapsedTime(self.events['decode'].start,
                                    self.events['decode'].end)[1]))
        print('RUN\t{:>9.2f} ms'.format(
            hip.hipEventElapsedTime(self.events['run'].start,
                                    self.events['run'].end)[1]))

    @staticmethod
    @measure
    def load_mgx_model(name,
                       shapes,
                       onnx_model_path,
                       compiled_model_path=None,
                       use_fp16=False,
                       force_compile=False,
                       exhaustive_tune=False,
                       offload_copy=True,
                       batch=1):
        print(f"Loading {name} model...")
        if compiled_model_path is None:
            compiled_model_path = onnx_model_path
        onnx_file = f"{onnx_model_path}/{name}/model.onnx"
        mxr_file = f"{compiled_model_path}/{name}/model_{'fp16' if use_fp16 else 'fp32'}_b{batch}_{'gpu' if not offload_copy else 'oc'}.mxr"
        if not force_compile and os.path.isfile(mxr_file):
            print(f"Found mxr, loading it from {mxr_file}")
            model = mgx.load(mxr_file, format="msgpack")
        elif os.path.isfile(onnx_file):
            print(f"No mxr found at {mxr_file}")
            print(f"Parsing from {onnx_file}")
            model = mgx.parse_onnx(onnx_file, map_input_dims=shapes)
            if use_fp16:
                mgx.quantize_fp16(model)
            model.compile(mgx.get_target("gpu"),
                          exhaustive_tune=exhaustive_tune,
                          offload_copy=offload_copy)
            print(f"Saving {name} model to {mxr_file}")
            os.makedirs(os.path.dirname(mxr_file), exist_ok=True)
            mgx.save(model, mxr_file, format="msgpack")
        else:
            print(
                f"No {name} model found at {onnx_file} or {mxr_file}. Please download it and re-try."
            )
            sys.exit(1)
        return model

    @measure
    def tokenize(self, prompt):
        return self.tokenizer.tokenize_with_weights(prompt)

    def encode_token_weights(self, model_name, token_weight_pairs):
        tokens = list(map(lambda a: a[0], token_weight_pairs[0]))
        tokens = torch.tensor([tokens], dtype=torch.int64, device=self.device)
        copy_tensor_sync(self.tensors[model_name]["input_ids"],
                         tokens.to(torch.int32))
        run_model_sync(self.models[model_name], self.model_args[model_name])
        encoder_out = self.tensors[model_name][get_output_name(0)]
        encoder_out2 = None
        if model_name != 't5xxl':
            # flipped outputs for clip text encoders...
            encoder_out2 = encoder_out
            encoder_out = self.tensors[model_name][get_output_name(1)]

        if encoder_out2 is not None:
            first_pooled = encoder_out2[0:1]
        else:
            first_pooled = encoder_out2
        output = [encoder_out[0:1]]

        return torch.cat(output, dim=-2), first_pooled

    @measure
    def get_embeddings(self, prompt_tokens):
        l_out, l_pooled = self.encode_token_weights("clip-l",
                                                    prompt_tokens["l"])
        # stable-diffusion-3-lite-onnx has swapped outputs for clip-l text encoder
        if l_out.shape != (1, 77, 768):
            l_out, l_pooled = l_pooled, l_out

        g_out, g_pooled = self.encode_token_weights("clip-g",
                                                    prompt_tokens["g"])
        if not self.skip_t5:
            t5_out, _ = self.encode_token_weights("t5xxl",
                                                  prompt_tokens["t5xxl"])
        else:
            t5_out = torch.zeros((1, 77, 4096)).cuda()
        lg_out = torch.cat([l_out, g_out], dim=-1)
        lg_out = torch.nn.functional.pad(lg_out, (0, 4096 - lg_out.shape[-1]))

        return torch.cat([lg_out, t5_out], dim=-2), torch.cat(
            (l_pooled, g_pooled), dim=-1)

    @staticmethod
    def convert_to_rgb_image(image):
        image = (image / 2 + 0.5).clamp(0, 1)
        image = image.detach().cpu().permute(0, 2, 3, 1).numpy()
        images = (image * 255).round().astype("uint8")
        return [Image.fromarray(images[i]) for i in range(images.shape[0])]

    @staticmethod
    def save_image(pil_image, filename="output.png"):
        pil_image.save(filename)

    def CFGDenoiser(self, x, timestep, cond, uncond, cond_scale):
        # Run cond and uncond in a batch together
        x_concat = torch.cat([x, x])
        timestep_concat = timestep.expand([2])
        c_crossattn = torch.cat([cond["c_crossattn"], uncond["c_crossattn"]])
        y = torch.cat([cond["y"], uncond["y"]])

        copy_tensor_sync(self.tensors["mmdit"]["hidden_states"], x_concat)
        copy_tensor_sync(self.tensors["mmdit"]["timestep"], timestep_concat)
        copy_tensor_sync(self.tensors["mmdit"]["encoder_hidden_states"],
                         c_crossattn)
        copy_tensor_sync(self.tensors["mmdit"]["pooled_projections"], y)

        run_model_sync(self.models["mmdit"], self.model_args['mmdit'])

        mmdit_out = self.tensors["mmdit"][get_output_name(0)]

        # Then split and apply CFG Scaling
        pos_out, neg_out = torch.tensor_split(mmdit_out, 2)

        scaled = neg_out + (pos_out - neg_out) * cond_scale

        # scheduler step function requies all tensors be on the CPU
        scaled = scaled.detach().clone().cpu()
        scheduler_out = self.scheduler.step(model_output=scaled,
                                            timestep=timestep,
                                            sample=x,
                                            return_dict=False)[0]
        return scheduler_out

    def fix_cond(self, cond):
        cond, pooled = (cond[0].cuda(), cond[1].cuda())
        return {"c_crossattn": cond, "y": pooled}

    def denoise(self, latent, conditioning, neg_cond, step, cfg_scale):
        conditioning = self.fix_cond(conditioning)
        neg_cond = self.fix_cond(neg_cond)
        return self.CFGDenoiser(latent, step, conditioning, neg_cond,
                                cfg_scale)

    @measure
    def decode(self, latents):
        copy_tensor_sync(self.tensors["vae"]["latent_sample"], latents)
        run_model_sync(self.models["vae"], self.model_args["vae"])
        return self.tensors["vae"][get_output_name(0)]

    @measure
    def warmup(self, num_runs):
        self.profile_start("warmup")
        copy_tensor_sync(self.tensors["clip-l"]["input_ids"],
                         torch.ones((1, 77)).to(torch.int32))
        copy_tensor_sync(self.tensors["clip-g"]["input_ids"],
                         torch.ones((1, 77)).to(torch.int32))
        if not self.skip_t5:
            copy_tensor_sync(self.tensors["t5xxl"]["input_ids"],
                             torch.ones((1, 77)).to(torch.int32))
        copy_tensor_sync(
            self.tensors["mmdit"]["hidden_states"],
            torch.randn((2 * self.batch, 16, 128, 128)).to(torch.float))
        copy_tensor_sync(self.tensors["mmdit"]["timestep"],
                         torch.randn((2 * self.batch)).to(torch.float))
        copy_tensor_sync(
            self.tensors["mmdit"]["encoder_hidden_states"],
            torch.randn((2 * self.batch, 154, 4096)).to(torch.float))
        copy_tensor_sync(self.tensors["mmdit"]["pooled_projections"],
                         torch.randn((2 * self.batch, 2048)).to(torch.float))
        copy_tensor_sync(
            self.tensors["vae"]["latent_sample"],
            torch.randn((self.batch, 16, 128, 128)).to(torch.float))

        for _ in range(num_runs):
            run_model_sync(self.models["clip-l"], self.model_args["clip-l"])
            run_model_sync(self.models["clip-g"], self.model_args["clip-g"])
            if not self.skip_t5:
                run_model_sync(self.models["t5xxl"], self.model_args["t5xxl"])
            run_model_sync(self.models["mmdit"], self.model_args["mmdit"])
            run_model_sync(self.models["vae"], self.model_args["vae"])
        self.profile_end("warmup")


if __name__ == "__main__":
    args = get_args()

    sd = StableDiffusionMGX(args.onnx_model_path, args.compiled_model_path,
                            args.fp16, args.batch, args.force_compile,
                            args.exhaustive_tune, args.skip_t5)
    print("Warmup")
    sd.warmup(5)
    print("Run")
    result = sd.run(args.prompt, args.negative_prompt, args.steps, args.seed,
                    args.scale)

    print("Summary")
    sd.print_summary(args.steps)
    print("Cleanup")
    sd.cleanup()

    print("Convert result to rgb image...")
    images = StableDiffusionMGX.convert_to_rgb_image(result)
    for i, image in enumerate(images):
        filename = f"{args.batch}_{args.output}" if args.output else f"output_s{args.seed}_t{args.steps}_{i}.png"
        StableDiffusionMGX.save_image(image, filename)
        print(f"Image saved to {filename}")



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import migraphx
import ctypes
import os
import glob


def test_conv_relu():
    # Full path of the library is needed to fix an issue on sles
    # where the library is not loaded otherwise.
    # We check for the presence of library at the following paths,
    # in the order listed below:
    #
    # 1. 'rocm_path' environment variable
    # 2. /opt/rocm
    # 3. /opt/rocm-*
    #
    # If the library is not found at any of these paths, we fall back
    # to the library path being detected automatically.

    library = "libamdhip64.so"

    # Environment variable containing path to rocm
    rocm_path_env_var = "rocm_path"

    # Check for rocm_path, default to /opt/rocm if it does not exist.
    rocm_path_var = os.getenv(rocm_path_env_var, default="/opt/rocm")

    # Join the paths to the library to get full path,
    # e.g. /opt/rocm/lib/libamdhip64.so
    library_file = os.path.join(rocm_path_var, "lib", library)

    # Check if the library file exists at the specified path
    if os.path.exists(library_file):
        # Replace library name by full path to the library
        library = library_file
    else:
        # Pattern match to look for path to different
        # rocm versions: /opt/rocm-*
        rocm_path_pattern = "/opt/rocm-*/lib/libamdhip64.so"
        matching_libraries = glob.glob(rocm_path_pattern)

        if matching_libraries:
            # Replace library name by full path to the first
            # library found.
            library = matching_libraries[0]

    # Loads library either by using the full path to the
    # library, if it has been detected earlier,
    # or, proceeds to load the library based on the name
    # of the library.
    hip = ctypes.cdll.LoadLibrary(library)

    p = migraphx.parse_onnx("conv_relu_maxpool_test.onnx")
    print(p)
    print("Compiling ...")
    # Need to have offload_copy = False to avoid syncs() back to the host device
    p.compile(migraphx.get_target("gpu"), offload_copy=False)
    print(p)
    params = {}

    # Using default value in api for hipSuccess which is always 0
    hipSuccess = ctypes.c_long(0)

    # Alloc a stream
    stream = ctypes.c_void_p()

    err = ctypes.c_long(
        hip.hipStreamCreateWithFlags(ctypes.byref(stream), ctypes.c_uint(0)))

    if err.value != hipSuccess.value:
        print("FAILED hipStreamCreate")
        return err

    # Use to_gpu to push generated argument to the GPU before we perform a run
    for key, value in p.get_parameter_shapes().items():
        params[key] = migraphx.to_gpu(migraphx.generate_argument(value))

    result = migraphx.from_gpu(
        p.run_async(params, stream.value, "ihipStream_t")[-1])

    # Wait for all commands in stream to complete
    err = ctypes.c_long(hip.hipStreamSynchronize(stream))
    if err.value != hipSuccess.value:
        print("FAILED: hipStreamSyncronize")
        return err

    # Cleanup Stream
    err = ctypes.c_long(hip.hipStreamDestroy(stream))
    if err.value != hipSuccess.value:
        print("FAILED: hipStreamDestroy")
        return err

    print(result)


test_conv_relu()



#  The MIT License (MIT)
#
#  Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the 'Software'), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from argparse import ArgumentParser
from diffusers import AutoencoderKL
import os
import torch


def argparser():
    parser = ArgumentParser()
    parser.add_argument(
        "-o",
        "--output_path",
        type=str,
        default="models/sdxl-1.0-base/vae_decoder_fp16_fix/model.onnx",
        help=
        "Path to save the onnx model. Use it to override the default models/<sdxl*> path."
    )
    return parser.parse_args()


class VAEDecoder(torch.nn.Module):
    def __init__(self, vae):
        super().__init__()
        self.vae = vae

    def forward(self, latent_sample):
        return self.vae.decode(latent_sample)


def export_vae_fp16(output_path):
    vae = AutoencoderKL.from_pretrained("madebyollin/sdxl-vae-fp16-fix")
    vae.eval()
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    torch.onnx.export(VAEDecoder(vae),
                      torch.randn(1, 4, 128, 128),
                      output_path,
                      export_params=True,
                      do_constant_folding=True,
                      input_names=['latent_sample'])


if __name__ == "__main__":
    args = argparser()
    export_vae_fp16(**vars(args))



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
import numpy as np
import json
import os.path
import tokenizers
import collections
from run_onnx_squad import read_squad_examples, write_predictions, convert_examples_to_features
import migraphx

RawResult = collections.namedtuple("RawResult",
                                   ["unique_id", "start_logits", "end_logits"])

#######################################
input_file = 'inputs_amd.json'
with open(input_file) as json_file:
    test_data = json.load(json_file)
    print(json.dumps(test_data, indent=2))

# preprocess input
predict_file = 'inputs_amd.json'

# Use read_squad_examples method from run_onnx_squad to read the input file
eval_examples = read_squad_examples(input_file=predict_file)

max_seq_length = 256
doc_stride = 128
max_query_length = 64
batch_size = 1
n_best_size = 20
max_answer_length = 30

vocab_file = os.path.join('uncased_L-12_H-768_A-12', 'vocab.txt')
tokenizer = tokenizers.BertWordPieceTokenizer(vocab_file)

# Use convert_examples_to_features method from run_onnx_squad to get parameters from the input
input_ids, input_mask, segment_ids, extra_data = convert_examples_to_features(
    eval_examples, tokenizer, max_seq_length, doc_stride, max_query_length)

#######################################
# Compile
print("INFO: Parsing and compiling the model...")
model = migraphx.parse_onnx("bertsquad-10.onnx")
model.compile(migraphx.get_target("gpu"))
#model.print()

print(model.get_parameter_names())
print(model.get_parameter_shapes())

n = len(input_ids)
bs = batch_size
all_results = []

for idx in range(0, n):
    item = eval_examples[idx]
    print(item)

    result = model.run({
        "unique_ids_raw_output___9:0":
        np.array([item.qas_id], dtype=np.int64),
        "input_ids:0":
        input_ids[idx:idx + bs],
        "input_mask:0":
        input_mask[idx:idx + bs],
        "segment_ids:0":
        segment_ids[idx:idx + bs]
    })

    in_batch = result[1].get_shape().lens()[0]
    start_logits = [float(x) for x in result[1].tolist()]
    end_logits = [float(x) for x in result[0].tolist()]
    for i in range(0, in_batch):
        unique_id = len(all_results)
        all_results.append(
            RawResult(unique_id=unique_id,
                      start_logits=start_logits,
                      end_logits=end_logits))

output_dir = 'predictions'
os.makedirs(output_dir, exist_ok=True)
output_prediction_file = os.path.join(output_dir, "predictions.json")
output_nbest_file = os.path.join(output_dir, "nbest_predictions.json")
write_predictions(eval_examples, extra_data, all_results, n_best_size,
                  max_answer_length, True, output_prediction_file,
                  output_nbest_file)

with open(output_prediction_file) as json_file:
    test_data = json.load(json_file)
    print(json.dumps(test_data, indent=2))



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#####################################################################################
from .base import BaseModel, DecoderModel, OptimumHFModelDownloadMixin
import numpy as np
from transformers import AutoTokenizer


class AutoTokenizerHFMixin(object):

    _processor = None

    @property
    def processor(self):
        if self._processor is None:
            self._processor = AutoTokenizer.from_pretrained(self.model_id)
        return self._processor

    def preprocess(self,
                   question,
                   context,
                   max_length,
                   truncation='only_second'):
        return self.processor(question,
                              context,
                              padding='max_length',
                              max_length=max_length,
                              truncation=truncation,
                              return_tensors="np")


class TextGenerationDecoderOnlyMixin(object):
    def __init__(self):
        # no pad token by default
        self.processor.pad_token = self.processor.eos_token

    def preprocess(self, *args, **kwargs):
        # swap squad's default "question - answer" order
        new_args, new_kwargs = list(args), kwargs
        new_args[0], new_args[1] = new_args[1], new_args[0]
        new_kwargs["truncation"] = "only_first"
        result = super().preprocess(*new_args, **new_kwargs)

        # result only contains "input_ids" and "attention_mask", extend it with "position_ids"
        result["position_ids"] = np.arange(0,
                                           len(result["input_ids"][0]),
                                           dtype=np.int64)
        result["position_ids"] = result["position_ids"][np.newaxis]
        return result

    def decode_step(self, input_map, output_map):
        timestep = np.argmax(
            input_map["input_ids"][0] == self.processor.eos_token_id)
        new_token = np.argmax(output_map["logits"][0][timestep - 1])
        input_map["input_ids"][0][timestep] = new_token
        input_map["attention_mask"][0][timestep] = 1
        return new_token == self.processor.eos_token_id


class BERT_large_uncased(OptimumHFModelDownloadMixin, AutoTokenizerHFMixin,
                         BaseModel):
    @property
    def model_id(self):
        return "google-bert/bert-large-uncased"

    @staticmethod
    def name():
        return "bert-large-uncased"

    def preprocess(self, *args, **kwargs):
        # use squad's question for masking
        question, context = args
        # replace first word with [MASK]
        masked_question = question.replace(question[:question.index(' ')],
                                           '[MASK]', 1)
        # swap question-answer order
        new_args = [context, masked_question]
        result = super().preprocess(*new_args, **kwargs)
        return result


class DistilBERT_base_cased_distilled_SQuAD(OptimumHFModelDownloadMixin,
                                            AutoTokenizerHFMixin, BaseModel):
    @property
    def model_id(self):
        return "distilbert/distilbert-base-cased-distilled-squad"

    @staticmethod
    def name():
        return "distilbert-base-cased-distilled-squad"


class RobertaBaseSquad2(OptimumHFModelDownloadMixin, AutoTokenizerHFMixin,
                        BaseModel):
    @property
    def model_id(self):
        return "deepset/roberta-base-squad2"

    @staticmethod
    def name():
        return "roberta-base-squad2"


# Note: the inheritance order here important
class GPTJ(OptimumHFModelDownloadMixin, TextGenerationDecoderOnlyMixin,
           AutoTokenizerHFMixin, DecoderModel):
    @property
    def model_id(self):
        return "EleutherAI/gpt-j-6b"

    @property
    def task(self):
        # override to ignore "with-past"
        return "text-generation"

    @staticmethod
    def name():
        return "gpt-j"


# Note: the inheritance order here important
class Llama2_7b_chat_hf(OptimumHFModelDownloadMixin,
                        TextGenerationDecoderOnlyMixin, AutoTokenizerHFMixin,
                        DecoderModel):
    @property
    def model_id(self):
        return "meta-llama/Llama-2-7b-chat-hf"

    @property
    def task(self):
        # override to ignore "with-past"
        return "text-generation"

    @staticmethod
    def name():
        return "llama2-7b-chat-hf"


# Note: the inheritance order here important
class Llama3_8b_instruct(OptimumHFModelDownloadMixin,
                         TextGenerationDecoderOnlyMixin, AutoTokenizerHFMixin,
                         DecoderModel):
    @property
    def model_id(self):
        return "meta-llama/Meta-Llama-3-8B-Instruct"

    @property
    def task(self):
        # override to ignore "with-past"
        return "text-generation"

    @staticmethod
    def name():
        return "llama3-8b-instruct"


class T5_base(OptimumHFModelDownloadMixin, AutoTokenizerHFMixin, DecoderModel):
    def __init__(self):
        max_length = 384  # default for squad
        # Note: no real start token, it requires the pad token
        self.initial_input_ids = np.array(
            [[self.processor.pad_token_id] + [self.processor.eos_token_id] *
             (max_length - 1)])

    @property
    def model_id(self):
        return "google-t5/t5-base"

    @property
    def task(self):
        # override to ignore "with-past"
        return "text2text-generation"

    @staticmethod
    def name():
        return "t5-base"

    def preprocess(self, *args, **kwargs):
        # only use squad's question
        new_args = ["translate English to French:", args[0]]
        result = super().preprocess(*new_args, **kwargs)
        result["decoder_input_ids"] = np.copy(self.initial_input_ids)
        return result

    def decode_step(self, input_map, output_map):
        timestep = np.argmax(
            input_map["decoder_input_ids"][0] == self.processor.eos_token_id)
        new_token = np.argmax(output_map["logits"][0][timestep - 1])
        input_map["decoder_input_ids"][0][timestep] = new_token
        input_map["attention_mask"][0][timestep] = 1
        return new_token == self.processor.eos_token_id


class Gemma_2b_it(OptimumHFModelDownloadMixin, AutoTokenizerHFMixin,
                  DecoderModel):
    @property
    def model_id(self):
        return "google/gemma-2b-it"

    @property
    def task(self):
        # override to ignore "with-past"
        return "text-generation"

    @staticmethod
    def name():
        return "gemma-2b-it"

    def preprocess(self, *args, **kwargs):
        # swap squad's default "question - answer" order
        new_args, new_kwargs = list(args), kwargs
        new_args[0], new_args[1] = new_args[1], new_args[0]
        new_kwargs["truncation"] = "only_first"
        result = super().preprocess(*new_args, **new_kwargs)
        # Note: padding-side is left, pos ids will be 1,1,1,...,0,1,2,3,...
        result["position_ids"] = np.cumsum(result["attention_mask"][0], -1) - 1
        result["position_ids"][:np.argmax(result["position_ids"] == 0)] = 1
        result["position_ids"] = result["position_ids"][np.newaxis]
        return result

    def decode_step(self, input_map, output_map):
        # The result is in the last logits
        new_token = np.argmax(output_map["logits"][0][-1])
        # Move everything left 1 step
        input_map["input_ids"][0] = np.roll(input_map["input_ids"][0], -1)
        input_map["input_ids"][0][-1] = new_token
        input_map["attention_mask"][0] = np.roll(
            input_map["attention_mask"][0], -1)
        input_map["attention_mask"][0][-1] = 1
        input_map["position_ids"][0] = np.roll(input_map["position_ids"][0],
                                               -1)
        input_map["position_ids"][0][-1] += input_map["position_ids"][0][-2]
        return new_token == self.processor.eos_token_id



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################
# This script generates tf pb files for MIGraphX tf operator tests.
# To generate an individual pb file, you can use the following
# command: python -c "import gen_tf_pb; gen_tf_pb.{test_name}_test()"
import tensorflow as tf


def tf_test(op_test):
    def run_test():
        g1 = tf.Graph()
        op_test(g1)
        tf.io.write_graph(g1,
                          './models',
                          '{}.pb'.format(op_test.__name__),
                          as_text=False)

    return run_test


@tf_test
def add_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='1')
        tf.add(g1_input, g2_input, name='add1')


@tf_test
def addv2_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='1')
        tf.raw_ops.AddV2(x=g1_input, y=g2_input, name='add1')


@tf_test
def add_bcast_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 3), name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 1), name='1')
        tf.math.add(g1_input, g2_input, name='add_bcast1')


@tf_test
def addn_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 3), name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 3), name='1')
        g3_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 3), name='2')

        tf.math.add_n([g1_input, g2_input, g3_input], name='addn1')        


@tf_test
def addn_single_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 3), name='0')

        tf.math.add_n([g1_input], name='addn1')


@tf_test
def argmax_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(3, 4, 5, 6),
                                            name='0')
        tf.argmax(g1_input, axis=2, name='argmax1')


@tf_test
def argmin_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(3, 4, 5, 6),
                                            name='0')
        tf.argmin(g1_input, axis=2, name='argmin1')


@tf_test
def assert_less_equal_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 3), name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 3), name='1')
        with tf.control_dependencies(
            [tf.compat.v1.assert_less_equal(g1_input, g2_input)]):
            tf.add(g1_input, g2_input, name='add1')


@tf_test
def batchmatmul_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 8, 4),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 4, 8),
                                            name='1')
        tf.matmul(g1_input,
                  g2_input,
                  transpose_a=True,
                  transpose_b=True,
                  name='batchmatmul1')


@tf_test
def batchnorm_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 32),
                                            name='x')
        g1_scale = tf.constant(1.0, dtype=tf.float32, shape=[32], name='scale')
        g1_offset = tf.compat.v1.placeholder(tf.float32,
                                             shape=(32),
                                             name='bias')
        g1_mean = tf.compat.v1.placeholder(tf.float32, shape=(32), name='mean')
        g1_variance = tf.compat.v1.placeholder(tf.float32,
                                               shape=(32),
                                               name='variance')
        tf.compat.v1.nn.fused_batch_norm(x=g1_input,
                                         scale=g1_scale,
                                         offset=g1_offset,
                                         mean=g1_mean,
                                         variance=g1_variance,
                                         epsilon=1e-4,
                                         is_training=False,
                                         name='batchnorm1')


@tf_test
def batchnorm_half_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float16,
                                            shape=(1, 16, 16, 32),
                                            name='x')
        g1_scale = tf.constant(1.0, dtype=tf.float32, shape=[32], name='scale')
        g1_offset = tf.compat.v1.placeholder(tf.float32,
                                             shape=(32),
                                             name='bias')
        g1_mean = tf.compat.v1.placeholder(tf.float32, shape=(32), name='mean')
        g1_variance = tf.compat.v1.placeholder(tf.float32,
                                               shape=(32),
                                               name='variance')
        tf.compat.v1.nn.fused_batch_norm(x=g1_input,
                                         scale=g1_scale,
                                         offset=g1_offset,
                                         mean=g1_mean,
                                         variance=g1_variance,
                                         epsilon=1e-4,
                                         is_training=False,
                                         name='batchnorm1')


@tf_test
def batchnormv3_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 32),
                                            name='x')
        g1_scale = tf.constant(1.0, dtype=tf.float32, shape=[32], name='scale')
        g1_offset = tf.compat.v1.placeholder(tf.float32,
                                             shape=(32),
                                             name='bias')
        g1_mean = tf.compat.v1.placeholder(tf.float32, shape=(32), name='mean')
        g1_variance = tf.compat.v1.placeholder(tf.float32,
                                               shape=(32),
                                               name='variance')
        tf.raw_ops.FusedBatchNormV3(x=g1_input,
                                    scale=g1_scale,
                                    offset=g1_offset,
                                    mean=g1_mean,
                                    variance=g1_variance,
                                    epsilon=1e-6,
                                    is_training=False,
                                    name='batchnorm1')


@tf_test
def biasadd_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 1, 1, 500),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32, shape=(500), name='1')
        tf.nn.bias_add(g1_input, g2_input, name='bias_add1')


@tf_test
def biasadd_scalar_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(1, 1), name='0')
        g2_const = tf.constant(1.0, tf.float32, shape=(1, ), name='1')
        tf.nn.bias_add(g1_input, g2_const, name='bias_add1')


@tf_test
def cast_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.cast(g1_input, dtype=tf.int32, name='cast1')


@tf_test
def concat_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(4, 7, 3),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(4, 2, 3),
                                            name='1')
        tf.concat([g1_input, g2_input], axis=1, name='concat1')


@tf_test
def const_test(g1):
    with g1.as_default():
        tf.constant(1.0, dtype=tf.float32, name='constant1')


@tf_test
def conv_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 3),
                                            name='0')
        g1_weights = tf.constant(value=1.0,
                                 dtype=tf.float32,
                                 shape=(3, 3, 3, 32),
                                 name='1')
        tf.nn.conv2d(g1_input, g1_weights, [1, 1, 1, 1], "SAME", name='conv1')


@tf_test
def conv_add_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 3),
                                            name='0')
        g1_weights = tf.constant(value=1.0,
                                 dtype=tf.float32,
                                 shape=(3, 3, 3, 32),
                                 name='1')
        conv = tf.nn.conv2d(g1_input,
                            g1_weights, [1, 1, 1, 1],
                            "SAME",
                            name='conv1')
        tf.add(conv, conv, name='add1')


@tf_test
def conv_batch_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(None, 16, 16, 3),
                                            name='0')
        g1_weights = tf.constant(value=1.0,
                                 dtype=tf.float32,
                                 shape=(3, 3, 3, 32),
                                 name='1')
        tf.nn.conv2d(g1_input, g1_weights, [1, 1, 1, 1], "SAME", name='conv1')


@tf_test
def conv_nchw_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        g1_weights = tf.constant(value=1.0,
                                 dtype=tf.float32,
                                 shape=(3, 3, 3, 32),
                                 name='1')
        tf.nn.conv2d(g1_input,
                     g1_weights, [1, 1, 1, 1],
                     "SAME",
                     data_format='NCHW',
                     name='conv1')


@tf_test
def conv_relu_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 3),
                                            name='0')
        g1_weights = tf.constant(value=1.0,
                                 dtype=tf.float32,
                                 shape=(3, 3, 3, 32),
                                 name='1')
        conv = tf.nn.conv2d(g1_input,
                            g1_weights, [1, 1, 1, 1],
                            "SAME",
                            name='conv1')
        tf.nn.relu(conv, name='relu1')


@tf_test
def conv_relu6_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 3),
                                            name='0')
        g1_weights = tf.constant(value=1.0,
                                 dtype=tf.float32,
                                 shape=(3, 3, 3, 32),
                                 name='1')
        conv = tf.nn.conv2d(g1_input,
                            g1_weights, [1, 1, 1, 1],
                            "SAME",
                            name='conv1')
        tf.nn.relu6(conv, name='relu1')


@tf_test
def depthwiseconv_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 3),
                                            name='0')
        g1_weights = tf.constant(value=1.0,
                                 dtype=tf.float32,
                                 shape=(3, 3, 3, 1),
                                 name='1')
        tf.compat.v1.nn.depthwise_conv2d_native(g1_input,
                                                g1_weights, [1, 1, 1, 1],
                                                "SAME",
                                                name='depthwiseconv1')


@tf_test
def expanddims_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(2, 3, 4),
                                            name='0')
        tf.expand_dims(g1_input, axis=0, name='expanddims_neg')


@tf_test
def gather_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 4), name='0')
        tf.gather(g1_input, [1, 1], axis=1, name='gather1')


@tf_test
def identity_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.identity(g1_input, 'identity')


@tf_test
def matmul_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(8, 4), name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32, shape=(4, 8), name='1')
        tf.matmul(g1_input,
                  g2_input,
                  transpose_a=True,
                  transpose_b=True,
                  name='matmul1')


@tf_test
def mean_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.math.reduce_mean(g1_input, axis=(2, 3), keepdims=True, name='mean1')
        tf.math.reduce_mean(g1_input,
                            axis=(2, 3),
                            keepdims=False,
                            name='mean2')


@tf_test
def mean_test_nhwc(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 3),
                                            name='0')
        tf.math.reduce_mean(g1_input,
                            axis=(1, 2),
                            keepdims=False,
                            name='mean2')


@tf_test
def mul_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 1, 1, 16),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 1, 1, 16),
                                            name='1')
        tf.multiply(g1_input, g2_input, name='mul1')


@tf_test
def multi_output_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.nn.relu(g1_input, 'relu')
        tf.tanh(g1_input, 'tanh')


@tf_test
def noop_test(g1):
    with g1.as_default():
        tf.raw_ops.NoOp(name='noop1')


@tf_test
def onehot_test(g1):
    with g1.as_default():
        g1_input = tf.constant((1, 1, 1, 1, 1), dtype=tf.int32)
        tf.one_hot(g1_input, 2, name='onehot1')


@tf_test
def pack_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(2), name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32, shape=(2), name='1')
        g3_input = tf.compat.v1.placeholder(tf.float32, shape=(2), name='2')
        tf.stack([g1_input, g2_input, g3_input], axis=1, name='pack1')


@tf_test
def pack_test_nhwc(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 1, 1, 2),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 1, 1, 2),
                                            name='1')
        g3_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 1, 1, 2),
                                            name='2')
        tf.stack([g1_input, g2_input, g3_input], axis=3, name='pack1')


@tf_test
def pad_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(2, 4), name='0')
        paddings = tf.constant([[1, 1], [2, 2]])

        tf.pad(g1_input, paddings, name='pad1')


@tf_test
def pooling_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 16, 16, 3),
                                            name='0')
        tf.compat.v1.nn.avg_pool(value=g1_input,
                                 ksize=(1, 2, 2, 1),
                                 strides=(1, 2, 2, 1),
                                 padding='VALID',
                                 data_format='NHWC',
                                 name='avg_pooling')
        tf.compat.v1.nn.max_pool(value=g1_input,
                                 ksize=(1, 2, 2, 1),
                                 strides=(1, 2, 2, 1),
                                 padding='VALID',
                                 data_format='NHWC',
                                 name='max_pooling')


@tf_test
def pow_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='1')
        tf.pow(g1_input, g2_input, name='pow1')


@tf_test
def relu_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.nn.relu(g1_input, 'relu')


@tf_test
def relu6_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.nn.relu6(g1_input, 'relu6')


@tf_test
def relu6_half_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float16,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.nn.relu6(g1_input, 'relu6')


@tf_test
def reshape_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(16), name='0')
        tf.reshape(g1_input, (1, 1, 1, 16), 'reshape')


@tf_test
def rsqrt_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.math.rsqrt(g1_input, 'rsqrt')


@tf_test
def shape_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
    g1.create_op(op_type='Shape', inputs=[g1_input])


@tf_test
def sigmoid_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.math.sigmoid(g1_input, 'sigmoid')


@tf_test
def slice_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(5, 10),
                                            name='0')
        tf.slice(g1_input, [1, 0], [2, -1], name='slice1')


@tf_test
def softmax_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32, shape=(1, 3), name='0')
        tf.nn.softmax(g1_input, name='softmax')


@tf_test
def split_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(5, 30),
                                            name='0')
        split0, split1, split2 = tf.split(g1_input, 3, 1, name='split')
        tf.concat([split0, split1], axis=1, name='concat1')
        tf.concat([split1, split2], axis=1, name='concat2')


@tf_test
def split_test_one_output(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(5, 30),
                                            name='0')
        tf.split(g1_input, 1, 1, name='split')


@tf_test
def split_test_vector_as_input(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(5, 30),
                                            name='0')
        split0, split1, split2 = tf.split(g1_input, [4, 15, 11],
                                          1,
                                          name='split')
        tf.concat([split0, split1], axis=1, name='concat1')
        tf.concat([split1, split2], axis=1, name='concat2')


@tf_test
def sqdiff_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='1')
        tf.compat.v1.squared_difference(g1_input, g2_input, name='sqdiff')


@tf_test
def squeeze_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 3, 1),
                                            name='0')
        tf.squeeze(g1_input, name='squeeze')


@tf_test
def stopgradient_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.stop_gradient(g1_input, 'stopgradient')


@tf_test
def stridedslice_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 1, 1, 10),
                                            name='0')
        tf.strided_slice(g1_input, [0, 0, 0, 0], [1, 1, 1, 5], [1, 1, 1, 1],
                         shrink_axis_mask=2,
                         name='stridedslice1')


@tf_test
def stridedslice_masks_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 3, 10),
                                            name='0')
        tf.strided_slice(g1_input, [0, 1, 1, 0], [0, 0, 0, 0], [1, 1, 1, 1],
                         begin_mask=9,
                         end_mask=15,
                         name='stridedslice1')


@tf_test
def sub_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='0')
        g2_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 2, 2, 3),
                                            name='1')
        tf.subtract(g1_input, g2_input, name='sub1')


@tf_test
def tanh_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.tanh(g1_input, 'tanh')


@tf_test
def transpose_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(1, 3, 16, 16),
                                            name='0')
        tf.transpose(g1_input, perm=[0, 2, 3, 1], name='transpose')


@tf_test
def variable_batch_test(g1):
    with g1.as_default():
        g1_input = tf.compat.v1.placeholder(tf.float32,
                                            shape=(0, 3, 16, 16),
                                            name='0')
        tf.identity(g1_input, name='identity')


if __name__ == '__main__':
    add_test()
    addv2_test()
    add_bcast_test()
    argmax_test()
    argmin_test()
    assert_less_equal_test()
    batchmatmul_test()
    batchnorm_test()
    batchnormv3_test()
    biasadd_test()
    biasadd_scalar_test()
    cast_test()
    concat_test()
    const_test()
    conv_test()
    conv_add_test()
    conv_batch_test()
    conv_nchw_test()
    conv_relu_test()
    conv_relu6_test()
    depthwiseconv_test()
    expanddims_test()
    gather_test()
    identity_test()
    matmul_test()
    mean_test()
    mean_test_nhwc()
    mul_test()
    multi_output_test()
    noop_test()
    onehot_test()
    pack_test()
    pack_test_nhwc()
    pad_test()
    pooling_test()
    pow_test()
    relu_test()
    relu6_test()
    relu6_half_test()
    reshape_test()
    rsqrt_test()
    shape_test()
    slice_test()
    softmax_test()
    split_test()
    split_test_one_output()
    split_test_vector_as_input()
    sqdiff_test()
    squeeze_test()
    stopgradient_test()
    stridedslice_test()
    stridedslice_masks_test()
    sub_test()
    tanh_test()
    transpose_test()
    variable_batch_test()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################

from txt2img import StableDiffusionMGX, get_args
import gradio as gr
import sys


class PrintWrapper(object):
    def __init__(self, org_handle):
        self.org_handle = org_handle
        self.log = ""

        def wrapper_write(x):
            self.log += x
            return org_handle.write(x)

        self.wrapper_write = wrapper_write

    def __getattr__(self, attr):
        return self.wrapper_write if attr == 'write' else getattr(
            self.org_handle, attr)

    def get_log(self):
        return self.log


def main():
    args = get_args()
    # Note: This will load the models, which can take several minutes
    sd = StableDiffusionMGX(args.pipeline_type, args.onnx_model_path,
                            args.compiled_model_path, args.use_refiner,
                            args.refiner_onnx_model_path,
                            args.refiner_compiled_model_path, args.fp16,
                            args.force_compile, args.exhaustive_tune)
    sd.warmup(5)

    def gr_wrapper(prompt, negative_prompt, steps, seed, scale, refiner_steps,
                   aesthetic_score, negative_aesthetic_score):
        img = None
        try:
            oldStdout, oldStderr = sys.stdout, sys.stderr
            sys.stdout, sys.stderr = PrintWrapper(sys.stdout), PrintWrapper(
                sys.stderr)
            result = sd.run(
                str(prompt),
                str(negative_prompt),
                int(steps),
                int(seed),
                float(scale),
                int(refiner_steps),
                float(aesthetic_score),
                float(negative_aesthetic_score),
                args.verbose,
            )
            img = StableDiffusionMGX.convert_to_rgb_image(result)
            sd.print_summary(steps)
        finally:
            log = ''.join([sys.stdout.get_log(), sys.stderr.get_log()])
            sys.stdout, sys.stderr = oldStdout, oldStderr
        return img, log

    demo = gr.Interface(gr_wrapper, [
        gr.Textbox(value=args.prompt, label="Prompt"),
        gr.Textbox(value=args.negative_prompt,
                   label="Negative prompt (Optional)"),
        gr.Slider(1, 100, step=1, value=args.steps, label="Number of steps"),
        gr.Textbox(value=args.seed, label="Random seed"),
        gr.Slider(1, 20, step=0.1, value=args.scale, label="Guidance scale"),
        gr.Slider(0,
                  100,
                  step=1,
                  value=args.refiner_steps,
                  label="Number of refiner steps. (Use 0 to skip it)",
                  visible=args.use_refiner),
        gr.Slider(1,
                  20,
                  step=0.1,
                  value=args.refiner_aesthetic_score,
                  label="Aesthetic score (Refiner)",
                  visible=args.use_refiner),
        gr.Slider(1,
                  20,
                  step=0.1,
                  value=args.refiner_negative_aesthetic_score,
                  label="Negative Aesthetic score (Refiner)",
                  visible=args.use_refiner),
    ], [
        "image",
        gr.Textbox(placeholder="Output log of the run", label="Output log")
    ])
    demo.launch()
    sd.cleanup()


if __name__ == "__main__":
    main()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################

import gradio as gr
from txtgen import Llama2MGX


def main():
    # Note: This will load the models, which can take several minutes
    llama = Llama2MGX(1024)

    def gr_wrapper(prompt):
        if prompt == "":
            return "Please provide a prompt."

        input_ids = llama.tokenize(prompt)
        result = llama.generate(input_ids)

        # trim input prompt from result
        result = result[len(prompt) + 2:]

        return result

    with gr.Blocks() as demo:
        gr.Markdown(
            "Start typing below and then click **Run** to see the output.")
        inp = gr.Textbox(placeholder="Type something here...",
                         label="Input prompt")
        btn = gr.Button("Run!")
        out = gr.Textbox(placeholder="The result will be displayed here",
                         label="Response")

        btn.click(fn=gr_wrapper, inputs=inp, outputs=out)

    demo.launch()


if __name__ == "__main__":
    main()



#####################################################################################
# The MIT License (MIT)
#
# Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#####################################################################################

from argparse import ArgumentParser
from transformers import LlamaTokenizer
import numpy as np
import migraphx as mgx
import os
import sys
import time
from functools import wraps


# measurement helper
def measure(fn):
    @wraps(fn)
    def measure_ms(*args, **kwargs):
        start_time = time.perf_counter_ns()
        result = fn(*args, **kwargs)
        end_time = time.perf_counter_ns()
        print(
            f"Elapsed time for {fn.__name__}: {(end_time - start_time) * 1e-6:.4f} ms\n"
        )
        return result

    return measure_ms


def get_args():
    parser = ArgumentParser()
    parser.add_argument(
        "-p",
        "--prompt",
        type=str,
        required=True,
        help="Input prompt",
    )

    parser.add_argument(
        "-l",
        "--log-process",
        action="store_true",
        help="Print the current state of transcribing.",
    )

    parser.add_argument("-s",
                        "--max-seq-len",
                        type=int,
                        choices=[256, 512, 1024, 2048, 4096],
                        default=1024,
                        help="Max sequence length the model can handle")

    return parser.parse_args()


class Llama2MGX():
    def __init__(self, max_seq_len=1024):
        model_id = "meta-llama/Llama-2-7b-chat-hf"
        self.max_seq_len = max_seq_len

        print("Load mgx model")
        self.model = Llama2MGX.load_mgx_model(
            max_seq_len, {
                "input_ids": [1, max_seq_len],
                "attention_mask": [1, max_seq_len],
                "position_ids": [1, max_seq_len]
            })
        print(f"Load AutoTokenizer model from {model_id}")
        self.tokenizer = LlamaTokenizer.from_pretrained(model_id)

    @staticmethod
    @measure
    def load_mgx_model(max_seq_len, shapes):
        file = "models/llama-2-7b-chat-hf/model"
        print(f"Loading {max_seq_len} seq-len version model from {file}")
        if os.path.isfile(f"{file}-{max_seq_len}.mxr"):
            print("Found mxr, loading it...")
            model = mgx.load(f"{file}-{max_seq_len}.mxr", format="msgpack")
        elif os.path.isfile(f"{file}.onnx"):
            print("Parsing from onnx file...")
            model = mgx.parse_onnx(f"{file}.onnx", map_input_dims=shapes)
            model.compile(mgx.get_target("gpu"))
            print("Saving model to mxr file...")
            mgx.save(model, f"{file}-{max_seq_len}.mxr", format="msgpack")
        else:
            print("No model found. Please download it and re-try.")
            sys.exit(1)
        return model

    @measure
    def tokenize(self, prompt):
        return self.tokenizer(prompt, return_tensors="np").input_ids

    @measure
    def get_input_features_for_input_ids(self, input_ids):
        input_ids_len = len(input_ids[0])
        padding_len = self.max_seq_len - input_ids_len
        input_ids = np.hstack([input_ids,
                               np.zeros((1, padding_len))]).astype(np.int64)
        # 0 masked | 1 un-masked
        attention_mask = np.array([1] * input_ids_len +
                                  [0] * padding_len).astype(np.int64)
        attention_mask = attention_mask[np.newaxis]
        position_ids = np.arange(0, self.max_seq_len, dtype=np.int64)
        position_ids = position_ids[np.newaxis]

        return (input_ids, attention_mask, position_ids)

    @measure
    def decode_step(self, input_ids, attention_mask, position_ids):
        return np.array(
            self.model.run({
                "input_ids": input_ids,
                "attention_mask": attention_mask,
                "position_ids": position_ids
            })[0])

    @measure
    def decode_tokens(self, generated_tokens, skip_special_tokens=True):
        return ''.join(
            self.tokenizer.decode(generated_tokens,
                                  skip_special_tokens=skip_special_tokens))

    @measure
    def generate(self, input_ids, log_process=False):
        start_timestep = len(input_ids[0]) - 1
        end_timestep = self.max_seq_len

        input_ids, attention_mask, position_ids = self.get_input_features_for_input_ids(
            input_ids)
        print("Generating response...")
        for timestep in range(start_timestep, self.max_seq_len):
            # get logits for the current timestep
            logits = self.decode_step(input_ids, attention_mask, position_ids)
            # greedily get the highest probable token
            new_token = np.argmax(logits[0][timestep])

            # add it to the tokens and unmask it
            input_ids[0][timestep + 1] = new_token
            attention_mask[0][timestep + 1] = 1

            if log_process:
                print(self.decode_tokens(input_ids[0][:timestep + 2]))

            if new_token == self.tokenizer.eos_token_id:
                end_timestep = timestep + 1
                break

        return self.decode_tokens(input_ids[0][:end_timestep + 1])


if __name__ == "__main__":
    args = get_args()
    llama = Llama2MGX(args.max_seq_len)

    print(f"Call tokenizer with \"{args.prompt}\"")

    input_ids = llama.tokenize(args.prompt)
    result = llama.generate(input_ids, log_process=args.log_process)

    print(f"Result text: {result}")



#  The MIT License (MIT)
#
#  Copyright (c) 2015-2024 Advanced Micro Devices, Inc. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the 'Software'), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from typing import List, Optional, Tuple
import argparse

import torch
import migraphx as mgx
import os
import sys

import torch.nn.functional as F

from rnnt_data import librespeech_huggingface

mgx_to_torch_dtype_dict = {
    "bool_type": torch.bool,
    "uint8_type": torch.uint8,
    "int8_type": torch.int8,
    "int16_type": torch.int16,
    "int32_type": torch.int32,
    "int64_type": torch.int64,
    "float_type": torch.float32,
    "double_type": torch.float64,
    "half_type": torch.float16,
}

torch_to_mgx_dtype_dict = {
    value: key
    for (key, value) in mgx_to_torch_dtype_dict.items()
}


def tensor_to_arg(tensor):
    return mgx.argument_from_pointer(
        mgx.shape(
            **{
                "type": torch_to_mgx_dtype_dict[tensor.dtype],
                "lens": list(tensor.size()),
                "strides": list(tensor.stride())
            }), tensor.data_ptr())


def tensors_to_args(tensors):
    return {name: tensor_to_arg(tensor) for name, tensor in tensors.items()}


def get_output_name(idx):
    return f"main:#output_{idx}"


def copy_tensor_sync(tensor, data):
    tensor.copy_(data)
    torch.cuda.synchronize()


def run_model_sync(model, args):
    model.run(args)
    mgx.gpu_sync()


def allocate_torch_tensors(model):
    input_shapes = model.get_parameter_shapes()
    data_mapping = {
        name: torch.zeros(shape.lens()).to(
            mgx_to_torch_dtype_dict[shape.type_string()]).to(device="cuda")
        for name, shape in input_shapes.items()
    }
    return data_mapping


class RNNT_MGX():
    def __init__(self, x, onnx_model_path, fp16):

        if fp16 is None:
            fp16 = []
        elif "all" in fp16:
            fp16 = ["rnnt_encoder", "rnnt_prediction", "rnnt_joint"]

        compiled_model_path = None
        force_compile = False
        exhaustive_tune = False
        print("Load models...")

        self.models = {
            "rnnt_encoder":
            RNNT_MGX.load_mgx_model(
                "rnnt_encoder",
                {
                    "input": x.shape,  # seq_length, batch_size, feature_length
                    "feature_length": [1]
                },
                onnx_model_path,
                compiled_model_path=compiled_model_path,
                use_fp16="rnnt_encoder" in fp16,
                force_compile=force_compile,
                exhaustive_tune=exhaustive_tune,
                offload_copy=False),
            "rnnt_prediction":
            RNNT_MGX.load_mgx_model("rnnt_prediction", {
                "symbol": [1, 1],
                "hidden_in_1": [2, 1, 320],
                "hidden_in_2": [2, 1, 320]
            },
                                    onnx_model_path,
                                    compiled_model_path=compiled_model_path,
                                    use_fp16="rnnt_prediction" in fp16,
                                    force_compile=force_compile,
                                    exhaustive_tune=exhaustive_tune,
                                    offload_copy=False),
            "rnnt_joint":
            RNNT_MGX.load_mgx_model("rnnt_joint", {
                "0": [1, 1, 1024],
                "1": [1, 1, 320]
            },
                                    onnx_model_path,
                                    compiled_model_path=compiled_model_path,
                                    use_fp16="rnnt_joint" in fp16,
                                    force_compile=force_compile,
                                    exhaustive_tune=exhaustive_tune,
                                    offload_copy=False)
        }

        self.tensors = {
            "rnnt_encoder":
            allocate_torch_tensors(self.models["rnnt_encoder"]),
            "rnnt_prediction":
            allocate_torch_tensors(self.models["rnnt_prediction"]),
            "rnnt_joint":
            allocate_torch_tensors(self.models["rnnt_joint"]),
        }

        self.model_args = {
            "rnnt_encoder": tensors_to_args(self.tensors['rnnt_encoder']),
            "rnnt_prediction":
            tensors_to_args(self.tensors['rnnt_prediction']),
            "rnnt_joint": tensors_to_args(self.tensors['rnnt_joint']),
        }

    @torch.no_grad()
    def encoder(self, inp, feature_length):
        copy_tensor_sync(self.tensors["rnnt_encoder"]["input"],
                         inp.to(torch.float32))
        copy_tensor_sync(self.tensors["rnnt_encoder"]["feature_length"],
                         feature_length.to(torch.int64))
        run_model_sync(self.models["rnnt_encoder"],
                       self.model_args["rnnt_encoder"])

        x_padded, x_lens = self.tensors["rnnt_encoder"][get_output_name(
            0)], self.tensors["rnnt_encoder"][get_output_name(1)]
        x_padded = x_padded.squeeze(1)
        x_padded = x_padded.transpose(1, 0)

        return x_padded.to(torch.float16), x_lens.to(torch.int32)

    @torch.no_grad()
    def prediction(self, symbol, hidden):
        copy_tensor_sync(self.tensors["rnnt_prediction"]["symbol"],
                         symbol.to(torch.int64))
        copy_tensor_sync(self.tensors["rnnt_prediction"]["hidden_in_1"],
                         hidden[0].to(torch.float32))
        copy_tensor_sync(self.tensors["rnnt_prediction"]["hidden_in_2"],
                         hidden[1].to(torch.float32))
        run_model_sync(self.models["rnnt_prediction"],
                       self.model_args["rnnt_prediction"])

        g = self.tensors["rnnt_prediction"][get_output_name(0)]
        hidden = self.tensors["rnnt_prediction"][get_output_name(1)].to(
            torch.float16), self.tensors["rnnt_prediction"][get_output_name(
                2)].to(torch.float16)

        return g.to(torch.float16), hidden

    @torch.no_grad()
    def joint(self, f, g):
        copy_tensor_sync(self.tensors["rnnt_joint"]["onnx::Shape_0"],
                         f.to(torch.float32))
        copy_tensor_sync(self.tensors["rnnt_joint"]["onnx::Shape_1"],
                         g.to(torch.float32))

        run_model_sync(self.models["rnnt_joint"],
                       self.model_args["rnnt_joint"])
        result = self.tensors["rnnt_joint"][get_output_name(0)]
        return result.to(torch.float16)

    @staticmethod
    def load_mgx_model(name,
                       shapes,
                       onnx_model_path,
                       compiled_model_path=None,
                       use_fp16=False,
                       force_compile=False,
                       exhaustive_tune=False,
                       offload_copy=True):
        print(f"Loading {name} model...")
        if compiled_model_path is None:
            compiled_model_path = onnx_model_path
        onnx_file = f"{onnx_model_path}/{name}/model.onnx"
        mxr_file = f"{compiled_model_path}/{name}/model_{'fp16' if use_fp16 else 'fp32'}_{'gpu' if not offload_copy else 'oc'}.mxr"
        if not force_compile and os.path.isfile(mxr_file):
            print(f"Found mxr, loading it from {mxr_file}")
            model = mgx.load(mxr_file, format="msgpack")
        elif os.path.isfile(onnx_file):
            print(f"No mxr found at {mxr_file}")
            print(f"Parsing from {onnx_file}")
            model = mgx.parse_onnx(onnx_file, map_input_dims=shapes)
            if use_fp16:
                mgx.quantize_fp16(model)
            model.compile(mgx.get_target("gpu"),
                          exhaustive_tune=exhaustive_tune,
                          offload_copy=offload_copy)
            print(f"Saving {name} model to {mxr_file}")
            os.makedirs(os.path.dirname(mxr_file), exist_ok=True)
            mgx.save(model, mxr_file, format="msgpack")
        else:
            print(
                f"No {name} model found at {onnx_file} or {mxr_file}. Please download it and re-try."
            )
            sys.exit(1)
        return model


class GreedyDecoder():
    def __init__(self, model, max_symbols_per_step=30):
        self._model = model
        self._SOS = -1
        self._blank_id = 28
        self._max_symbols_per_step = max_symbols_per_step

    def run(self, x: torch.Tensor, out_lens: torch.Tensor
            ) -> Tuple[torch.Tensor, torch.Tensor, List[List[int]]]:
        logits, logits_lens = self._model.encoder(x, out_lens)

        output: List[List[int]] = []
        for batch_idx in range(logits.size(0)):
            inseq = logits[batch_idx, :, :].unsqueeze(1).to('cuda')
            # inseq: TxBxF
            logitlen = logits_lens[batch_idx]
            sentence = self._greedy_decode(inseq, logitlen)
            output.append(sentence)

        return logits, logits_lens, output

    def _greedy_decode(self, x: torch.Tensor,
                       out_len: torch.Tensor) -> List[int]:
        hidden: Optional[Tuple[torch.Tensor, torch.Tensor]] = None
        label: List[int] = []
        for time_idx in range(int(out_len.item())):
            f = x[time_idx, :, :].unsqueeze(0).to('cuda')

            not_blank = True
            symbols_added = 0

            while not_blank and symbols_added < self._max_symbols_per_step:

                g, hidden_prime = self._pred_step(self._get_last_symb(label),
                                                  hidden)

                logp = self._joint_step(f, g, log_normalize=False)[0, :]

                # get index k, of max prob
                v, k = logp.max(0)
                k = k.item()

                if k == self._blank_id:
                    not_blank = False
                else:
                    label.append(k)
                    hidden = hidden_prime

                symbols_added += 1

        return label

    def _pred_step(self, label: int,
                   hidden: Optional[Tuple[torch.Tensor, torch.Tensor]]
                   ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        if label == self._SOS:
            label = 0

        label = torch.tensor([[label]], dtype=torch.int64)

        if hidden is None:
            hidden = torch.zeros([2, 1, 320]).to(device="cuda"), torch.zeros(
                [2, 1, 320]).to(device="cuda")

        g, hidden = self._model.prediction(label, hidden)

        return g, hidden

    def _joint_step(self,
                    enc: torch.Tensor,
                    pred: torch.Tensor,
                    log_normalize: bool = False) -> torch.Tensor:
        logits = self._model.joint(enc, pred)[:, 0, 0, :]

        if not log_normalize:
            return logits

        probs = F.log_softmax(logits, dim=len(logits.shape) - 1)

        return probs

    def _get_last_symb(self, labels: List[int]) -> int:
        return self._SOS if len(labels) == 0 else labels[-1]


def decode_string(result):
    string = ''
    for c in result[0]:
        if c == 0:
            string += " "
        else:
            string += chr(c + 96)
    return string


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--onnx_model_path', default='models/rnnt/')
    parser.add_argument(
        "--fp16",
        choices=["all", "rnnt_encoder", "rnnt_prediction", "rnnt_joint"],
        nargs="+",
        help="Quantize models with fp16 precision.",
    )

    args = parser.parse_args()

    model_parts = ['rnnt_encoder', 'rnnt_joint', 'rnnt_prediction']
    check_onnx_files = [
        os.path.isfile(f"{args.onnx_model_path}/{name}/model.onnx")
        for name in model_parts
    ]

    print("Getting data...")
    x, out_lens, transcript = librespeech_huggingface()

    if any(check_onnx_files) == False:
        from rnnt_torch_model import pytorch_rnnt_model
        from rnnt_onnx import export_rnnt_onnx

        print("Create pytorch model...")
        pytorch_model = pytorch_rnnt_model()

        print("Export pytorch model to ONNX...")
        export_rnnt_onnx(pytorch_model, x, args.onnx_model_path)

    print("Read MIGX model from ONNX and run...")
    migx_model = RNNT_MGX(x,
                          onnx_model_path=args.onnx_model_path,
                          fp16=args.fp16)
    rnnt_decoder = GreedyDecoder(migx_model)
    _, _, result = rnnt_decoder.run(x.to(torch.float32), out_lens)
    print("Transcribed Sentence: ", decode_string(result))
    print("Ground Truth: ", transcript)
